###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
Contains verdi commands for fleurinpdata
"""
import click
from aiida.cmdline.commands.cmd_data.cmd_list import query, list_options
from aiida.cmdline.params import arguments, options, types
from aiida.cmdline.utils import decorators, echo
from aiida.cmdline.params.types import DataParamType
from aiida.plugins import DataFactory
#from aiida_fleur.data.fleurinp import FleurinpData
from . import cmd_data

FleurinpData = DataFactory('fleur.fleurinp')


@click.group('fleurinp')
def cmd_fleurinp():
    """Commands to handle `FleurinpData` nodes."""


cmd_data.add_command(cmd_fleurinp)


@cmd_fleurinp.command('list')
@list_options  # usual aiida list options
@click.option('--uuid/--no-uuid', default=False, show_default=True, help='Display uuid of nodes.')
@click.option('--ctime/--no-ctime', default=False, show_default=True, help='Display ctime of nodes.')
@click.option('--extras/--no-extras', default=True, show_default=True, help='Display extras of nodes.')
@click.option('--strucinfo/--no-strucinfo',
              default=False,
              show_default=True,
              help='Perpare additional information on the crystal structure to show. This slows down the query.')
@decorators.with_dbenv()
def list_fleurinp(raw, past_days, groups, all_users, strucinfo, uuid, ctime, extras):
    """
    List stored FleurinpData in the database with additional information
    """
    # do a query and list all reuse AiiDA code
    from tabulate import tabulate
    list_project_headers = ['Id', 'Label', 'Description', 'Files']  # these we always get
    # 'UUID', 'Ctime',
    columns_dict = {
        'ID': 'id',
        'Id': 'id',
        'UUID': 'uuid',
        'Ctime': 'ctime',
        'Label': 'label',
        'Description': 'description',
        'Files': 'attributes.files',
        'Extras': 'attributes.extras'
    }

    if uuid:
        list_project_headers.append('UUID')
    if ctime:
        list_project_headers.append('Ctime')
    if extras:
        list_project_headers.append('Extras')

    project = [columns_dict[k] for k in list_project_headers]
    group_pks = None
    if groups is not None:
        group_pks = [g.pk for g in groups]

    data_fleurinp = query(FleurinpData, project, past_days, group_pks, all_users)
    if strucinfo:  # second query
        # we get the whole node to get some extra information
        project2 = '*'
        fleurinps = query(FleurinpData, project2, past_days, group_pks, all_users)
        list_project_headers.append('Formula')
    counter = 0
    fleurinp_list_data = []

    # , 'Formula', 'Symmetry'
    # It is fastest for list commands to only display content from a query
    if not raw:
        fleurinp_list_data.append(list_project_headers)
    for j, entry in enumerate(data_fleurinp):
        #print(entry)
        for i, value in enumerate(entry):
            if isinstance(value, list):
                new_entry = []
                for elm in value:
                    if elm is None:
                        new_entry.append('')
                    else:
                        new_entry.append(elm)
                entry[i] = ','.join(new_entry)
        if strucinfo:
            structure = fleurinps[j][0].get_structuredata_ncf()
            formula = structure.get_formula()
            entry.append(formula)
        for i in range(len(entry), len(list_project_headers)):
            entry.append(None)
        counter += 1
    fleurinp_list_data.extend(data_fleurinp)
    if raw:
        echo.echo(tabulate(fleurinp_list_data, tablefmt='plain'))
    else:
        echo.echo(tabulate(fleurinp_list_data, headers='firstrow'))
        echo.echo(f'\nTotal results: {counter}\n')


@cmd_fleurinp.command('cat')
@arguments.NODE('node', type=DataParamType(sub_classes=('aiida.data:fleur.fleurinp',)))
@click.option('-f',
              '--filename',
              'filename',
              default='inp.xml',
              show_default=True,
              help='Disply the file content of the given filename.')
def cat_file(node, filename):
    """
    Dumb the content of a file contained in given fleurinpdata, per default dump
    inp.xml
    """
    echo.echo(node.get_content(filename=filename))


'''
@cmd_fleurinp.command('info')
def info():
    """
    Shows some basic information about the fleurinp datastructure and dumbs the
    inp.xml
    """
    click.echo('Not implemented yet, sorry. Please implement me!')


@cmd_fleurinp.command('show')
def cmd_show():
    """
    Shows the content of a certain file
    """
    click.echo('Not implemented yet, sorry. Please implement me!')


@cmd_fleurinp.command('open')
def open_inp():
    """
    opens the inp.xml in some editor, readonly.
    inp.xml this way looking at xml might be more convenient.
    """
    click.echo('Not implemented yet, sorry. Please implement me!')


# this is a maybe
@cmd_fleurinp.command()
def get_structure():
    """
    Prints some basic information about the structure data and return a structure uuid/pk
    """
    click.echo('Not implemented yet, sorry. Please implement me!')


@cmd_fleurinp.command()
def get_kpoints():
    """
    Prints some basic information about the kpoints data and returns a kpoints uuid/pk
    """
    click.echo('Not implemented yet, sorry. Please implement me!')


@cmd_fleurinp.command()
def get_parameters():
    """
    Prints some basic information about the parameter data and returns a
    parameter data uuid/pk
    """
    click.echo('Not implemented yet, sorry. Please implement me!')
'''



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
    In this module you find the workflow 'FleurCFCoeffWorkChain' for calculating
    the 4f crystal field coefficients
"""
from aiida.engine import WorkChain, ToContext, ExitCode
from aiida.engine import calcfunction as cf
from aiida.common import AttributeDict
from aiida.common.exceptions import NotExistent
from aiida import orm
from aiida.common.constants import elements as PeriodicTableElements

from aiida_fleur.tools.StructureData_util import replace_element, mark_atoms, get_atomtype_site_symmetry
from aiida_fleur.tools.common_fleur_wf import get_inputs_fleur
from aiida_fleur.data.fleurinpmodifier import FleurinpModifier
from aiida_fleur.calculation.fleur import FleurCalculation

from aiida_fleur.workflows.scf import FleurScfWorkChain
from aiida_fleur.workflows.base_fleur import FleurBaseWorkChain
from aiida_fleur.workflows.orbcontrol import FleurOrbControlWorkChain

from masci_tools.tools.cf_calculation import CFCalculation, CFCoefficient
from masci_tools.util.schema_dict_util import eval_simple_xpath, tag_exists

import h5py
from lxml import etree
import numpy as np


def reconstruct_cfcoeffcients(output_dict, atomtype=None):
    """
    Reconstruct the CFCoefficient list from the output dictionary
    of the FleurCFCoeffWorkChain

    :param output_dict: output dictionary node or the corresponding
                        dictionary
    :param atomtype: int of the atomtype to reconstruct the coefficients for
    """
    if isinstance(output_dict, orm.Dict):
        output_dict = output_dict.get_dict()

    if atomtype is not None and not isinstance(atomtype, str):
        atomtype = str(atomtype)

    multiple_atomtypes = all('/' not in key for key in output_dict['cf_coefficients_spin_up'])

    if atomtype is None and multiple_atomtypes:
        raise ValueError('atomtype not specified')

    if multiple_atomtypes and atomtype not in output_dict['cf_coefficients_spin_up']:
        raise ValueError(f'Atomtype {atomtype} not available')

    spin_up = output_dict['cf_coefficients_spin_up']
    spin_down = output_dict['cf_coefficients_spin_down']
    spin_up_imag = output_dict.get('cf_coefficients_spin_up_imag', {})
    spin_down_imag = output_dict.get('cf_coefficients_spin_down_imag', {})

    convention = output_dict['cf_coefficients_convention']
    unit = output_dict['cf_coefficients_units']

    if multiple_atomtypes:
        spin_up = spin_up[atomtype]
        spin_down = spin_down[atomtype]
        spin_up_imag = spin_up_imag.get(atomtype, {})
        spin_down_imag = spin_down_imag.get(atomtype, {})

    coefficients = []
    for key in spin_up:
        l, m = key.split('/', maxsplit=1)
        up = spin_up[key]
        if key in spin_up_imag:
            up = up + 1j * spin_up_imag[key]
        down = spin_down[key]
        if key in spin_down_imag:
            down = down + 1j * spin_down_imag[key]

        coefficients.append(CFCoefficient(l=l, m=m, spin_up=up, spin_down=down, unit=unit, convention=convention))

    return coefficients


def reconstruct_cfcalculation(charge_densities, potentials, atomtype, **kwargs):
    """
    Reconstruct the CFCalculation instance from the outputs of the
    FleurCFCoeffWorkChain
    """

    radial_meshes = {'cdn': charge_densities.get_x()[1], 'pot': potentials.get_x()[1]}
    names, cdn_array, _ = zip(*charge_densities.get_y())
    if f'atomtype-{atomtype}' not in names:
        raise ValueError(f'Atomtype {atomtype} not available')
    density = cdn_array[names.index(f'atomtype-{atomtype}')]

    names, pot_array, _ = zip(*potentials.get_y())

    pot_dict = {}
    for l in range(0, 7):
        for m in range(-l, l + 1):
            if f'atomtype-{atomtype}-{l}/{m}-up' in names:
                p = [pot_array[names.index(f'atomtype-{atomtype}-{l}/{m}-up')]]
                if f'atomtype-{atomtype}-{l}/{m}-down' in names:
                    p.append(pot_array[names.index(f'atomtype-{atomtype}-{l}/{m}-down')])
                pot_dict[(l, m)] = np.array(p)

    cfcalc = CFCalculation.from_arrays(density, potentials=pot_dict, radial_mesh=radial_meshes, **kwargs)

    return cfcalc


class FleurCFCoeffWorkChain(WorkChain):
    """
    Workflow for calculating rare-earth crystal field coefficients
    """
    _workflowversion = '0.2.0'

    _wf_default = {
        'element': '',
        'rare_earth_analogue': False,
        'analogue_element': 'Y',
        'replace_all': True,
        'soc_off': True,
        'convert_to_stevens': True,
    }

    _CF_GROUP_LABEL = '89999'

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.expose_inputs(FleurScfWorkChain,
                           namespace='scf_rare_earth_analogue',
                           exclude=('structure', 'fleurinp'),
                           namespace_options={
                               'required': False,
                               'populate_defaults': False
                           })
        spec.expose_inputs(FleurScfWorkChain,
                           namespace='scf',
                           namespace_options={
                               'required': False,
                               'populate_defaults': False
                           })
        spec.expose_inputs(FleurOrbControlWorkChain,
                           namespace='orbcontrol',
                           namespace_options={
                               'required': False,
                               'populate_defaults': False
                           })
        spec.input('wf_parameters', valid_type=orm.Dict, required=False)

        spec.outline(cls.start, cls.validate_input, cls.run_scfcalculations, cls.run_cfcalculation, cls.return_results)

        spec.output('output_cfcoeff_wc_para', valid_type=orm.Dict)
        spec.output('output_cfcoeff_wc_charge_densities', valid_type=orm.XyData, required=False)
        spec.output('output_cfcoeff_wc_potentials', valid_type=orm.XyData, required=False)

        spec.exit_code(230, 'ERROR_INVALID_INPUT_PARAM', message='Invalid workchain parameters.')
        spec.exit_code(231, 'ERROR_INVALID_INPUT_CONFIG', message='Invalid input configuration.')
        spec.exit_code(235, 'ERROR_CHANGING_FLEURINPUT_FAILED', message='Input file modification failed.')
        spec.exit_code(236, 'ERROR_INVALID_INPUT_FILE', message="Input file was corrupted after user's modifications.")
        spec.exit_code(345, 'ERROR_SCF_FAILED', message='Convergence scf workflow failed.')
        spec.exit_code(451, 'ERROR_ORBCONTROL_FAILED', message='Convergence orbcontrol workflow failed.')
        spec.exit_code(452, 'ERROR_CFCALC_FAILED', message='CF calculation failed.')

    def start(self):
        """
        init context and some parameters
        """
        self.report(f'INFO: started crystal field coefficient workflow version {self._workflowversion}')

        ####### init    #######

        # internal para /control para
        self.ctx.successful = True
        self.ctx.info = []
        self.ctx.warnings = []
        self.ctx.errors = []
        self.ctx.num_analogues = None
        wf_default = self._wf_default
        if 'wf_parameters' in self.inputs:
            wf_dict = self.inputs.wf_parameters.get_dict()
        else:
            wf_dict = wf_default

        for key, val in wf_default.items():
            wf_dict[key] = wf_dict.get(key, val)
        self.ctx.wf_dict = wf_dict

    def validate_input(self):
        """
        validate input
        """
        extra_keys = {key for key in self.ctx.wf_dict if key not in self._wf_default}
        if extra_keys:
            error = f'ERROR: input wf_parameters for CFCoeff contains extra keys: {extra_keys}'
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        inputs = self.inputs
        if 'scf' not in inputs and 'orbcontrol' not in inputs:
            error = 'ERROR: Missing input. Provide one of the scf or orbcontrol inputs.'
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
        elif 'scf' in inputs and 'orbcontrol' in inputs:
            error = 'ERROR: Invalid Input. Provide only one of the scf or orbcontrol inputs.'
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_CONFIG

        element = self.ctx.wf_dict['element']
        atomic_numbers = {data['symbol']: num for num, data in PeriodicTableElements.items()}
        if element not in atomic_numbers:
            error = f'ERROR: Invalid Input. Element not a valid element: {element}'
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_PARAM
        else:
            if atomic_numbers[element] < 57 and atomic_numbers[element] > 70:
                error = 'ERROR: Invalid Input. CF coefficient workflow only implemented for 4f rare-earths'
                self.report(error)
                return self.exit_codes.ERROR_INVALID_INPUT_PARAM

    def run_scfcalculations(self):

        self.report('INFO: Starting SCF calculations')
        inputs = {}
        calcs = {}
        if self.ctx.wf_dict['rare_earth_analogue']:
            self.report(f"INFO: Creating Rare-Earth Analogue with {self.ctx.wf_dict['analogue_element']}")
            all_inputs = self.get_inputs_rare_earth_analogue()
            for name, inputs in all_inputs.items():
                calcs[name] = self.submit(FleurScfWorkChain, **inputs)
                calcs[name].label = name
                calcs[
                    name].description = f"SCF workflow for the rare-earth analogue ({self.ctx.wf_dict['analogue_element']}); number {name.split('_')[-1]}"

        if 'scf' in self.inputs:
            inputs = self.get_inputs_scf()
            result_scf = self.submit(FleurScfWorkChain, **inputs)
            calcs['rare_earth_scf'] = result_scf
            calcs['rare_earth_scf'].label = 'rare_earth_scf'
            calcs['rare_earth_scf'].description = 'SCF workflow for the rare-earth system for the CF calculation'
        elif 'orbcontrol' in self.inputs:
            inputs = self.get_inputs_orbcontrol()
            result_orbcontrol = self.submit(FleurOrbControlWorkChain, **inputs)
            calcs['rare_earth_orbcontrol'] = result_orbcontrol
            calcs['rare_earth_orbcontrol'].label = 'rare_earth_orbcontrol'
            calcs[
                'rare_earth_orbcontrol'].description = 'Orbcontrol workflow for the rare-earth system for the CF calculation'

        return ToContext(**calcs)

    def get_inputs_rare_earth_analogue(self):

        inputs = self.inputs
        if 'scf' in inputs:
            input_scf = AttributeDict(self.exposed_inputs(FleurScfWorkChain, namespace='scf'))
        elif 'orbcontrol' in self.inputs:
            input_scf = AttributeDict(self.exposed_inputs(FleurOrbControlWorkChain, namespace='orbcontrol'))
            input_scf_tmp = input_scf.get('scf_no_ldau')
            if input_scf_tmp is None:
                input_scf_tmp = {}
                input_scf_tmp['structure'] = input_scf['structure']
                input_scf_tmp['inpgen'] = input_scf['inpgen']
                input_scf_tmp['fleur'] = input_scf['fleur']
                input_scf_tmp['options'] = input_scf['options']
                if 'calc_parameters' in input_scf:
                    input_scf_tmp['calc_parameters'] = input_scf['calc_parameters']
            input_scf = input_scf_tmp

        if 'structure' in input_scf:
            orig_structure = input_scf['structure']
        elif 'fleurinp' in input_scf:
            orig_structure = input_scf['fleurinp'].get_structuredata_ncf()

        if 'calc_parameters' in input_scf:
            rare_earth_params = input_scf['calc_parameters'].get_dict()
        elif 'fleurinp' in input_scf:
            rare_earth_params = input_scf['fleurinp'].get_parameterdata_ncf(write_ids=False).get_dict()
        else:
            rare_earth_params = {}

        replace_dict = {}
        replace_dict[self.ctx.wf_dict['element']] = self.ctx.wf_dict['analogue_element']

        new_structures = replace_element(orig_structure,
                                         orm.Dict(dict=replace_dict),
                                         replace_all=orm.Bool(self.ctx.wf_dict['replace_all']))

        inputs = {}
        self.ctx.num_analogues = len(new_structures)
        for index, structure in enumerate(new_structures.values()):
            inputs_analogue = AttributeDict(self.exposed_inputs(FleurScfWorkChain, namespace='scf_rare_earth_analogue'))
            inputs_analogue.structure = structure
            # inputs_analogue.structure = mark_atoms(structure,
            #                                        lambda _, kind: kind.symbols ==
            #                                        (self.ctx.wf_dict['analogue_element'],),
            #                                        kind_id=self._CF_GROUP_LABEL)

            if 'calc_parameters' not in inputs_analogue:

                #Reuse parameters from rare earth calculation
                new_params = rare_earth_params.copy()
                for key, value in rare_earth_params.items():
                    if 'atom' in key:
                        if 'element' in value:
                            if value['element'] == self.ctx.wf_dict['element']:
                                new_params.pop(key)
                inputs_analogue.calc_parameters = orm.Dict(dict=new_params)

            if self.ctx.wf_dict['soc_off']:
                if 'wf_parameters' not in inputs_analogue:
                    scf_wf_dict = {}
                else:
                    scf_wf_dict = inputs_analogue.wf_parameters.get_dict()

                scf_wf_dict.setdefault('inpxml_changes', []).append(('set_species', {
                    'attributedict': {
                        'special': {
                            'socscale': 0.0
                        }
                    },
                    'species_name':
                    f"all-{self.ctx.wf_dict['analogue_element']}"
                }))

            inputs_analogue.wf_parameters = orm.Dict(dict=scf_wf_dict)
            inputs_analogue.metadata.call_link_label = f'analogue_scf_{index}'
            inputs[f'analogue_scf_{index}'] = inputs_analogue

        return inputs

    def get_inputs_scf(self):

        input_scf = AttributeDict(self.exposed_inputs(FleurScfWorkChain, namespace='scf'))

        if self.ctx.wf_dict['soc_off']:
            if 'wf_parameters' not in input_scf:
                scf_wf_dict = {}
            else:
                scf_wf_dict = input_scf.wf_parameters.get_dict()

            scf_wf_dict.setdefault('inpxml_changes', []).append(('set_species', {
                'species_name': f"all-{self.ctx.wf_dict['element']}",
                'attributedict': {
                    'special': {
                        'socscale': 0.0
                    }
                }
            }))

            input_scf.wf_parameters = orm.Dict(dict=scf_wf_dict)
        input_scf.metadata.call_link_label = 'rare_earth_scf'

        return input_scf

    def get_inputs_orbcontrol(self):

        input_orbcontrol = AttributeDict(self.exposed_inputs(FleurOrbControlWorkChain, namespace='orbcontrol'))

        if self.ctx.wf_dict['soc_off'] and 'scf_no_ldau' in input_orbcontrol:
            if 'wf_parameters' not in input_orbcontrol['scf_no_ldau']:
                scf_wf_dict = {}
            else:
                scf_wf_dict = input_orbcontrol['scf_no_ldau'].wf_parameters.get_dict()

            scf_wf_dict.setdefault('inpxml_changes', []).append(('set_species', {
                'species_name': f"all-{self.ctx.wf_dict['element']}",
                'attributedict': {
                    'special': {
                        'socscale': 0.0
                    }
                }
            }))

            input_orbcontrol.scf_no_ldau.wf_parameters = orm.Dict(dict=scf_wf_dict)
        elif self.ctx.wf_dict['soc_off']:
            if 'wf_parameters' not in input_orbcontrol:
                orbcontrol_wf_dict = {}
            else:
                orbcontrol_wf_dict = input_orbcontrol.wf_parameters.get_dict()

            orbcontrol_wf_dict.setdefault('inpxml_changes', []).append(('set_species', {
                'species_name': f"all-{self.ctx.wf_dict['element']}",
                'attributedict': {
                    'special': {
                        'socscale': 0.0
                    }
                }
            }))

            input_orbcontrol.wf_parameters = orm.Dict(dict=orbcontrol_wf_dict)
        input_orbcontrol.metadata.call_link_label = 'rare_earth_orbcontrol'

        return input_orbcontrol

    def run_cfcalculation(self):

        if 'scf' in self.inputs:
            if not self.ctx.rare_earth_scf.is_finished_ok:
                error = ('ERROR: SCF workflow (rare-earth) was not successful')
                self.report(error)
                return self.exit_codes.ERROR_SCF_FAILED

            try:
                outdict = self.ctx.rare_earth_scf.outputs.output_scf_wc_para
            except NotExistent:
                message = ('ERROR: SCF workflow (rare-earth) failed, no scf output node')
                self.ctx.errors.append(message)
                return self.exit_codes.ERROR_SCF_FAILED
        else:
            if not self.ctx.rare_earth_orbcontrol.is_finished_ok:
                if self.ctx.rare_earth_orbcontrol.exit_status not in FleurOrbControlWorkChain.get_exit_statuses(
                    ['ERROR_SOME_CONFIGS_FAILED']):
                    error = ('ERROR: Orbcontrol workflow (rare-earth) was not successful')
                    self.report(error)
                    return self.exit_codes.ERROR_ORBCONTROL_FAILED

            try:
                outdict = self.ctx.rare_earth_orbcontrol.outputs.output_orbcontrol_wc_para
            except NotExistent:
                message = ('ERROR: Orbcontrol workflow (rare-earth) failed, no orbcontrol output node')
                self.ctx.errors.append(message)
                self.report(message)
                return self.exit_codes.ERROR_ORBCONTROL_FAILED

            try:
                outdict = self.ctx.rare_earth_orbcontrol.outputs.groundstate_scf.output_scf_wc_para
            except NotExistent:
                message = ('ERROR: Orbcontrol workflow (rare-earth) failed, no groundstate scf output node')
                self.ctx.errors.append(message)
                self.report(message)
                return self.exit_codes.ERROR_ORBCONTROL_FAILED

        if self.ctx.wf_dict['rare_earth_analogue']:
            if not all(self.ctx[f'analogue_scf_{index}'].is_finished_ok for index in range(self.ctx.num_analogues)):
                error = (
                    f"ERROR: One SCF workflow ({self.ctx.wf_dict['analogue_element']}-analogue) was not successful")
                self.report(error)
                return self.exit_codes.ERROR_SCF_FAILED

            try:
                for index in range(self.ctx.num_analogues):
                    _ = self.ctx[f'analogue_scf_{index}'].outputs.output_scf_wc_para
            except NotExistent:
                message = (
                    f"ERROR: SCF workflow ({self.ctx.wf_dict['analogue_element']}-analogue) failed, no scf output node")
                self.ctx.errors.append(message)
                self.report(message)
                return self.exit_codes.ERROR_SCF_FAILED

        self.report('INFO: Running Crystal Field Calculations')
        calcs = {}
        if self.ctx.wf_dict['rare_earth_analogue']:
            all_inputs = self.get_inputs_cfanalogue_calculation()
            for name, inputs in all_inputs.items():
                calcs[name] = self.submit(FleurBaseWorkChain, **inputs)
                calcs[name].label = name
                calcs[
                    name].description = f"Calculation of crystal field potential with {self.ctx.wf_dict['analogue_element']} Analogue Method"

        inputs = self.get_inputs_cfrareearth_calculation()
        result_rareearth = self.submit(FleurBaseWorkChain, **inputs)
        calcs['rare_earth_cf'] = result_rareearth
        calcs['rare_earth_cf'].label = 'rare_earth_cf'
        calcs['rare_earth_cf'].description = 'Crystal Field Calculation including the 4f element'

        return ToContext(**calcs)

    def get_inputs_cfanalogue_calculation(self):

        analogue_element = self.ctx.wf_dict['analogue_element']

        all_inputs = {}
        for index in range(self.ctx.num_analogues):
            inputs = AttributeDict(self.exposed_inputs(FleurScfWorkChain, namespace='scf_rare_earth_analogue'))

            fleurinp_scf = self.ctx[f'analogue_scf_{index}'].outputs.fleurinp
            remote_data = self.ctx[f'analogue_scf_{index}'].outputs.last_calc.remote_folder

            if 'settings' in inputs:
                settings = inputs.settings.get_dict()
            else:
                settings = {}

            if 'options' in inputs:
                options = inputs.options.get_dict()
            else:
                options = {}

            fm = FleurinpModifier(fleurinp_scf)

            fm.set_atomgroup(attributedict={'cFCoeffs': {
                'chargeDensity': False,
                'potential': True
            }},
                             species=f"all-{self.ctx.wf_dict['analogue_element']}")

            try:
                fm.show(display=False, validate=True)
            except etree.DocumentInvalid:
                error = ('ERROR: input, inp.xml changes did not validate')
                self.control_end_wc(error)
                return {}, self.exit_codes.ERROR_INVALID_INPUT_FILE
            except ValueError as exc:
                error = ('ERROR: input, inp.xml changes could not be applied.\n'
                         f'The following error was raised {exc}')
                self.control_end_wc(error)
                return {}, self.exit_codes.ERROR_CHANGING_FLEURINPUT_FAILED

            fleurinp_cf = fm.freeze()

            label = f'analogue_cf_{index}'
            description = f'Calculation of crystal field potential with {analogue_element} Analogue Method'

            all_inputs[label] = get_inputs_fleur(inputs.fleur,
                                                 remote_data,
                                                 fleurinp_cf,
                                                 options,
                                                 label,
                                                 description,
                                                 settings=settings)
            all_inputs[label].setdefault('metadata', {})['call_link_label'] = label

        return all_inputs

    def get_inputs_cfrareearth_calculation(self):

        if 'scf' in self.inputs:
            inputs = AttributeDict(self.exposed_inputs(FleurScfWorkChain, namespace='scf'))

            fleurinp_scf = self.ctx.rare_earth_scf.outputs.fleurinp
            remote_data = self.ctx.rare_earth_scf.outputs.last_calc.remote_folder
        elif 'orbcontrol' in self.inputs:
            inputs = AttributeDict(self.exposed_inputs(FleurOrbControlWorkChain, namespace='orbcontrol'))

            fleurinp_scf = self.ctx.rare_earth_orbcontrol.outputs.groundstate_scf.fleurinp
            remote_data = self.ctx.rare_earth_orbcontrol.outputs.groundstate_scf.last_calc.remote_folder

        if 'settings' in inputs:
            settings = inputs.settings.get_dict()
        else:
            settings = {}

        if 'options' in inputs:
            options = inputs.options.get_dict()
        else:
            options = {}

        label = 'rare_earth_cf'
        if self.ctx.wf_dict['rare_earth_analogue']:
            description = 'Calculation of crystal field charge density'
        else:
            description = 'Calculation of crystal field potential/charge density'

        fm = FleurinpModifier(fleurinp_scf)
        element = self.ctx.wf_dict['element']
        if self.ctx.wf_dict['rare_earth_analogue']:
            #Only charge density
            fm.set_atomgroup(attributedict={'cFCoeffs': {
                'chargeDensity': True,
                'potential': False
            }},
                             species=f'all-{element}')
        else:
            #Both potential and charge density
            fm.set_atomgroup(attributedict={'cFCoeffs': {
                'chargeDensity': True,
                'potential': True,
                'remove4f': True
            }},
                             species=f'all-{element}')

        try:
            fm.show(display=False, validate=True)
        except etree.DocumentInvalid:
            error = ('ERROR: input, inp.xml changes did not validate')
            self.control_end_wc(error)
            return {}, self.exit_codes.ERROR_INVALID_INPUT_FILE
        except ValueError as exc:
            error = ('ERROR: input, inp.xml changes could not be applied.\n'
                     f'The following error was raised {exc}')
            self.control_end_wc(error)
            return {}, self.exit_codes.ERROR_CHANGING_FLEURINPUT_FAILED

        fleurinp_cf = fm.freeze()

        inputs_rareearth = get_inputs_fleur(inputs.fleur,
                                            remote_data,
                                            fleurinp_cf,
                                            options,
                                            label,
                                            description,
                                            settings=settings)
        inputs_rareearth.setdefault('metadata', {})['call_link_label'] = label

        return inputs_rareearth

    def check_cf_calculation(self, calc_name):
        """
        Check that the CFCalculation finished successfully
        """
        successful = True

        if calc_name in self.ctx:
            calc = self.ctx[calc_name]
        else:
            message = f'One CF calculation was not run because the scf workflow failed: {calc_name}'
            self.ctx.warnings.append(message)
            successful = False
            self.ctx.successful = self.ctx.successful and successful
            return successful

        if not calc.is_finished_ok:
            message = f'One CF calculation was not successful: {calc_name}'
            self.ctx.warnings.append(message)
            successful = False

        try:
            _ = calc.outputs.output_parameters
        except NotExistent:
            message = f'One CF calculation failed, no output node: {calc_name}. I skip this one.'
            self.ctx.errors.append(message)
            successful = False

        if FleurCalculation._CFDATA_HDF5_FILE_NAME not in calc.outputs.retrieved.list_object_names():
            message = f'One CF calculation did not produce a {FleurCalculation._CFDATA_HDF5_FILE_NAME} file: {calc_name}'
            self.ctx.warnings.append(message)
            successful = False

        self.ctx.successful = self.ctx.successful and successful
        return successful

    def return_results(self):
        """
        Return results fo cf calculation
        """

        outnodedict = {}
        cf_calcs_out = []  #All single nodes
        charge_densities = []
        potentials = []

        #This calculation is always there
        success = self.check_cf_calculation('rare_earth_cf')
        if success:
            link_label = 'rare_earth_cf'
            outnodedict[link_label] = self.ctx.rare_earth_cf.outputs.output_parameters
            cdn_retrieved = self.ctx.rare_earth_cf.outputs.retrieved
            xmltree, schema_dict = self.ctx.rare_earth_cf.inputs.fleurinpdata.load_inpxml()

            groups = eval_simple_xpath(xmltree, schema_dict, 'atomGroup', list_return=True)
            atomTypes = []
            for index, group in enumerate(groups):
                if tag_exists(group, schema_dict, 'cfcoeffs'):
                    atomTypes.append(index + 1)

            if not self.ctx.wf_dict['rare_earth_analogue']:
                pot_retrieved = self.ctx.rare_earth_cf.outputs.retrieved
                res = calculate_cf_coefficients(cdn_retrieved,
                                                pot_retrieved,
                                                convert=orm.Bool(self.ctx.wf_dict['convert_to_stevens']),
                                                atomTypes=orm.List(list=atomTypes))
                if isinstance(res, ExitCode):
                    self.report(f'Calculation of crystal field coefficients failed with {cf_calc_out!r}')
                    self.ctx.successful = False
                    success = False
                    cf_calc_out = {}
                else:
                    cf_calc_out = res['out']
                    cf_calcs_out = [cf_calc_out]
                    cf_calc_out = cf_calc_out.get_dict()
                    charge_densities = [res['charge_densities']]
                    potentials = [res['potentials']]
            else:
                cf_calc_out = {}
                for index in range(self.ctx.num_analogues):
                    calc_name = f'analogue_cf_{index}'
                    success = self.check_cf_calculation(calc_name)
                    if not success:
                        continue
                    pot_retrieved = self.ctx[calc_name].outputs.retrieved
                    outnodedict[link_label] = self.ctx[calc_name].outputs.output_parameters

                    xmltree, schema_dict = self.ctx[calc_name].inputs.fleurinpdata.load_inpxml()
                    groups = eval_simple_xpath(xmltree, schema_dict, 'atomGroup', list_return=True)
                    atomTypes = []
                    for group_index, group in enumerate(groups):
                        if tag_exists(group, schema_dict, 'cfcoeffs'):
                            atomTypes.append(group_index + 1)

                    cf_calc_out_analogue = calculate_cf_coefficients(cdn_retrieved,
                                                                     pot_retrieved,
                                                                     convert=orm.Bool(
                                                                         self.ctx.wf_dict['convert_to_stevens']),
                                                                     atomTypes=orm.List(list=atomTypes))
                    if isinstance(cf_calc_out_analogue, ExitCode):
                        self.report(
                            f'Calculation of crystal field coefficients failed: {calc_name} with {cf_calc_out!r}')
                        self.ctx.successful = False
                        continue

                    charge_densities.append(cf_calc_out_analogue['charge_densities'])
                    potentials.append(cf_calc_out_analogue['potentials'])
                    cf_calc_out_analogue = cf_calc_out_analogue['out']

                    cf_calcs_out.append(cf_calc_out_analogue)

                    if not cf_calc_out:
                        cf_calc_out = cf_calc_out_analogue.get_dict()
                    else:
                        cf_calc_out['cf_coefficients_atomtypes'] += atomTypes

                    cf_calc_out['cf_coefficients_spin_up'] = {
                        **cf_calc_out['cf_coefficients_spin_up'],
                        **cf_calc_out_analogue['cf_coefficients_spin_up']
                    }
                    cf_calc_out['cf_coefficients_spin_down'] = {
                        **cf_calc_out['cf_coefficients_spin_down'],
                        **cf_calc_out_analogue['cf_coefficients_spin_down']
                    }
                    if not self.ctx.wf_dict['convert_to_stevens']:
                        cf_calc_out['cf_coefficients_spin_up_imag'] = {
                            **cf_calc_out['cf_coefficients_spin_up_imag'],
                            **cf_calc_out_analogue['cf_coefficients_spin_up_imag']
                        }
                        cf_calc_out['cf_coefficients_spin_down_imag'] = {
                            **cf_calc_out['cf_coefficients_spin_down_imag'],
                            **cf_calc_out_analogue['cf_coefficients_spin_down_imag']
                        }

        #pop out output if only one atomtype is calculated
        if success and len(cf_calc_out['cf_coefficients_atomtypes']) == 1:
            _, cf_calc_out['cf_coefficients_spin_up'] = cf_calc_out['cf_coefficients_spin_up'].popitem()
            _, cf_calc_out['cf_coefficients_spin_down'] = cf_calc_out['cf_coefficients_spin_down'].popitem()
            if not self.ctx.wf_dict['convert_to_stevens']:
                _, cf_calc_out['cf_coefficients_spin_up_imag'] = cf_calc_out['cf_coefficients_spin_up_imag'].popitem()
                _, cf_calc_out['cf_coefficients_spin_down_imag'] = cf_calc_out[
                    'cf_coefficients_spin_down_imag'].popitem()

        rare_earth_site_symmetries = []
        if success and len(cf_calc_out['cf_coefficients_atomtypes']) > 0:
            #For this to work the order of the atomtype CANNOT change between the conversions
            struc = self.ctx.rare_earth_cf.inputs.fleurinpdata.get_structuredata_ncf()
            site_symmetries = get_atomtype_site_symmetry(struc)
            rare_earth_site_symmetries = [
                site_symmetries[atomtype - 1] for atomtype in cf_calc_out['cf_coefficients_atomtypes']
            ]

        out = {
            'workflow_name': self.__class__.__name__,
            'workflow_version': self._workflowversion,
            'angle_a_to_x_axis': None,
            'angle_c_to_z_axis': None,
            'density_normalization': None,
            'cf_coefficients_spin_up': None,
            'cf_coefficients_spin_down': None,
            'cf_coefficients_convention': None,
            'cf_coefficients_units': 'K',
            'cf_coefficients_site_symmetries': rare_earth_site_symmetries,
            'info': self.ctx.info,
            'warnings': self.ctx.warnings,
            'errors': self.ctx.errors
        }

        if cf_calc_out:
            for key, value in cf_calc_out.items():
                out[key] = value

        if self.ctx.successful:
            self.report('Done, Crystal Field coefficients calculation complete')
        else:
            self.report('Done, but something went wrong.... Probably some individual calculation failed or'
                        ' a scf-cycle did not reach the desired distance or the post-processing failed.')

        outnode = orm.Dict(dict=out)
        outnodedict = {f'crystal_field_coefficients_{index}': cf_coff for index, cf_coff in enumerate(cf_calcs_out)}
        outnodedict['results_node'] = outnode

        if charge_densities:
            #Merge the different calculations together
            cdn_output = orm.XyData()
            x_name, x_array, x_unit = charge_densities[0].get_x()
            cdn_output.set_x(x_array, x_name, x_unit)

            y_names, y_arrays, y_units = [], [], []
            for cdn in charge_densities:
                names, arrays, units = zip(*cdn.get_y())
                y_names.extend(names)
                y_arrays.extend(arrays)
                y_units.extend(units)
            cdn_output.set_y(y_arrays, y_names, y_units)

            cdn_output.label = 'output_cfcoeff_wc_charge_densities'
            cdn_output.description = 'Charge densities used in the Crystal Field calculation'
            outnodedict['charge_densities'] = cdn_output

            #Merge the different calculations together
            pot_output = orm.XyData()
            x_name, x_array, x_unit = potentials[0].get_x()
            pot_output.set_x(x_array, x_name, x_unit)

            y_names, y_arrays, y_units = [], [], []
            for pot in potentials:
                names, arrays, units = zip(*pot.get_y())
                y_names.extend(names)
                y_arrays.extend(arrays)
                y_units.extend(units)
            pot_output.set_y(y_arrays, y_names, y_units)

            pot_output.label = 'output_cfcoeff_wc_potentials'
            pot_output.description = 'Potentials used in the Crystal Field calculation'
            outnodedict['potentials'] = pot_output

        # create links between all these nodes...
        outputnode_dict = create_cfcoeff_results_node(**outnodedict)
        outputnode = outputnode_dict.get('output_cfcoeff_wc_para')
        outputnode.label = 'output_cfcoeff_wc_para'
        outputnode.description = (
            'Contains crystal field occupation results and information of an FleurCFCoeffWorkChain run.')

        returndict = {}
        returndict['output_cfcoeff_wc_para'] = outputnode
        if charge_densities:
            returndict['output_cfcoeff_wc_charge_densities'] = cdn_output
            returndict['output_cfcoeff_wc_potentials'] = pot_output

        # create link to workchain node
        for link_name, node in returndict.items():
            self.out(link_name, node)

        if not self.ctx.successful:
            return self.exit_codes.ERROR_CFCALC_FAILED

    def control_end_wc(self, errormsg):
        """
        Controlled way to shutdown the workchain. It will initialize the output nodes
        The shutdown of the workchain will has to be done afterwards
        """
        self.ctx.successful = False
        self.report(errormsg)
        self.ctx.errors.append(errormsg)
        self.return_results()


@cf
def create_cfcoeff_results_node(**kwargs):
    """
    This is a pseudo cf, to create the right graph structure of AiiDA.
    This calcfunction will create the output nodes in the database.
    It also connects the output_nodes to all nodes the information comes from.
    This includes the output_parameter node for the orbcontrol, connections to run scfs,
    and returning of the gs_calculation (best initial density matrix)
    So far it is just parsed in as kwargs argument, because we are to lazy
    to put most of the code overworked from return_results in here.
    """
    outdict = {}
    outpara = kwargs.get('results_node', {})
    outdict['output_cfcoeff_wc_para'] = outpara.clone()

    return outdict


@cf
def calculate_cf_coefficients(cf_cdn_folder: orm.FolderData,
                              cf_pot_folder: orm.FolderData,
                              convert: orm.Bool = None,
                              atomTypes: orm.List = None) -> orm.Dict:
    """
    Calculate the crystal filed coefficients using the tool from the
    masci-tools package

    :param cf_cdn_folder: FolderData for the retrieved files for the charge density data
    :param cf_pot_folder: FolderData for the retrieved files for the potential data

    :raises: ExitCode 300, CFData.hdf file is missing
    :raises: ExitCode 310, CFdata.hdf reading failed
    :raises: ExitCode 320, Crystal field calculation failed
    """

    if convert is None:
        convert = orm.Bool(True)

    out_dict = {}

    units = None
    convention = None
    phi = None
    theta = None
    norm = None
    coefficients_dict_up = {}
    coefficients_dict_dn = {}
    coefficients_dict_up_imag = {}
    coefficients_dict_dn_imag = {}

    charge_densities = {}
    charge_densities_rmesh = None
    potentials = {}
    potentials_rmesh = None

    if atomTypes is None:
        res = _calculate_single_atomtype(cf_cdn_folder, cf_pot_folder, convert)
        if isinstance(res, ExitCode):
            return res
        cfcalc, coefficients = res
        atomTypes = orm.List(list=[cfcalc.atom_type])
        for coeff in coefficients:
            if units is None:
                units = coeff.unit
                convention = coeff.convention
            key = f'{coeff.l}/{coeff.m}'
            coefficients_dict_up[key] = coeff.spin_up.real
            coefficients_dict_dn[key] = coeff.spin_down.real
            coefficients_dict_up_imag[key] = coeff.spin_up.imag
            coefficients_dict_dn_imag[key] = coeff.spin_down.imag
        phi = cfcalc.phi
        theta = cfcalc.theta
        norm = cfcalc.density_normalization

        atom_prefix = f'atomtype-{cfcalc.atom_type}'
        charge_densities_rmesh, charge_densities[atom_prefix] = cfcalc.get_charge_density(interpolated=False)

        potentials_rmesh, potentials_up = cfcalc.get_potentials('up', interpolated=False)
        potentials_down = {}
        if cfcalc.spin_polarized:
            _, potentials_down = cfcalc.get_potentials('down', interpolated=False)

        for (l, m), pot in potentials_up.items():
            prefix = f'{prefix}-{l}/{m}'
            potentials[f'{prefix}-up'] = pot
            if (l, m) in potentials_down:
                potentials[f'{prefix}-down'] = potentials_down[(l, m)]

    else:
        norm = {}
        for atom_type in atomTypes:
            res = _calculate_single_atomtype(cf_cdn_folder, cf_pot_folder, convert, atom_type=atom_type)
            if isinstance(res, ExitCode):
                return res
            cfcalc, coefficients = res

            atom_up = coefficients_dict_up.setdefault(atom_type, {})
            atom_dn = coefficients_dict_dn.setdefault(atom_type, {})
            atom_up_imag = coefficients_dict_up_imag.setdefault(atom_type, {})
            atom_dn_imag = coefficients_dict_dn_imag.setdefault(atom_type, {})
            norm[atom_type] = cfcalc.density_normalization
            phi = cfcalc.phi
            theta = cfcalc.theta

            for coeff in coefficients:
                if units is None:
                    units = coeff.unit
                    convention = coeff.convention
                key = f'{coeff.l}/{coeff.m}'
                atom_up[key] = coeff.spin_up.real
                atom_dn[key] = coeff.spin_down.real
                atom_up_imag[key] = coeff.spin_up.imag
                atom_dn_imag[key] = coeff.spin_down.imag

            atom_prefix = f'atomtype-{atom_type}'
            rmesh, charge_densities[atom_prefix] = cfcalc.get_charge_density(interpolated=False)
            if charge_densities_rmesh is None:
                charge_densities_rmesh = rmesh

            rmesh, potentials_up = cfcalc.get_potentials('up', interpolated=False)
            if potentials_rmesh is None:
                potentials_rmesh = rmesh
            potentials_down = {}
            if cfcalc.spin_polarized:
                _, potentials_down = cfcalc.get_potentials('down', interpolated=False)

            for (l, m), pot in potentials_up.items():
                prefix = f'{atom_prefix}-{l}/{m}'
                potentials[f'{prefix}-up'] = pot
                if (l, m) in potentials_down:
                    potentials[f'{prefix}-down'] = potentials_down[(l, m)]

    out_dict['cf_coefficients_atomtypes'] = atomTypes.get_list()
    out_dict['cf_coefficients_spin_up'] = coefficients_dict_up
    out_dict['cf_coefficients_spin_down'] = coefficients_dict_dn
    if not convert:
        out_dict['cf_coefficients_spin_up_imag'] = coefficients_dict_up_imag
        out_dict['cf_coefficients_spin_down_imag'] = coefficients_dict_dn_imag
    #Output more information about the performed calculation
    out_dict['angle_a_to_x_axis'] = phi
    out_dict['angle_c_to_z_axis'] = theta
    out_dict['density_normalization'] = norm
    out_dict['cf_coefficients_units'] = units
    out_dict['cf_coefficients_convention'] = convention

    out_dict = orm.Dict(dict=out_dict)
    out_dict.label = 'CFCoefficients'
    out_dict.description = 'Results of the post-processing tool for calculating Crystal field coefficients'

    cdn_data = orm.XyData()
    cdn_data.set_x(charge_densities_rmesh, 'rmesh', x_units='Bohr')

    y_names, y_arrays = zip(*charge_densities.items())
    y_units = ['density'] * len(y_names)
    cdn_data.set_y(y_arrays, y_names, y_units=y_units)

    cdn_data.label = 'cfcoeff_cdn_data'
    cdn_data.description = ('Contains XyData for the Charge density used in the crystal field calculation')

    pot_data = orm.XyData()
    pot_data.set_x(potentials_rmesh, 'rmesh', x_units='Bohr')

    y_names, y_arrays = zip(*potentials.items())
    y_units = ['htr'] * len(y_names)
    pot_data.set_y(y_arrays, y_names, y_units=y_units)

    pot_data.label = 'cfcoeff_pot_data'
    pot_data.description = ('Contains XyData for the Poteintials used in the crystal field calculation')

    return {'out': out_dict, 'charge_densities': cdn_data, 'potentials': pot_data}


def _calculate_single_atomtype(cf_cdn_folder, cf_pot_folder, convert, **kwargs):
    """
    Private method wrapping the calculation of coefficients
    """
    CRYSTAL_FIELD_FILE = FleurCalculation._CFDATA_HDF5_FILE_NAME

    cfcalc = CFCalculation(quiet=True)
    #Reading in the HDF files
    if CRYSTAL_FIELD_FILE in cf_cdn_folder.list_object_names():
        try:
            with cf_cdn_folder.open(CRYSTAL_FIELD_FILE, 'rb') as f:
                with h5py.File(f, 'r') as cffile:
                    cfcalc.read_charge_density(cffile, **kwargs)
        except ValueError as exc:
            return ExitCode(310, message=f'{CRYSTAL_FIELD_FILE} reading failed with: {exc}')
    else:
        return ExitCode(300, message=f'{CRYSTAL_FIELD_FILE} file not in the retrieved files')

    if CRYSTAL_FIELD_FILE in cf_pot_folder.list_object_names():
        try:
            with cf_pot_folder.open(CRYSTAL_FIELD_FILE, 'rb') as f:
                with h5py.File(f, 'r') as cffile:
                    cfcalc.read_potential(cffile, **kwargs)
        except ValueError as exc:
            return ExitCode(310, message=f'{CRYSTAL_FIELD_FILE} reading failed with: {exc}')
    else:
        return ExitCode(300, message=f'{CRYSTAL_FIELD_FILE} file not in the retrieved files')

    try:
        coefficients = cfcalc.get_coefficients(convention='Stevens' if convert else 'Wybourne')
    except ValueError as exc:
        return ExitCode(320, message=f'Crystal field calculation failed with: {exc}')
    if len(coefficients) == 0:
        return ExitCode(320, message='Crystal field calculation failed with: No Coefficients produced')

    return cfcalc, coefficients



# -*- coding: utf-8 -*-
from aiida.orm import load_node, Dict
from aiida.engine import submit

from aiida_fleur.workflows.create_magnetic_film import FleurCreateMagneticWorkChain

fleur_code = load_node(FLEUR_PK)
inpgen_code = load_node(INPGEN_PK)

wf_para = {
    'lattice': 'fcc',
    'directions': [[-1, 1, 0],
               [0, 0, 1],
               [1, 1, 0]],
    'host_symbol': 'Pt',
    'latticeconstant': 4.0,
    'size': (1, 1, 5),
    'replacements': {0: 'Fe', -1: 'Fe'},
    'decimals': 10,
    'pop_last_layers': 1,

    'total_number_layers': 8,
    'num_relaxed_layers': 3,
}

wf_para = Dict(dict=wf_para)

wf_eos = {'points': 15,
          'step': 0.015,
          'guess': 1.00
          }

wf_eos_scf = {'fleur_runmax': 4,
              'density_converged': 0.0002,
              'itmax_per_run': 50,
              'inpxml_changes': []
              }

wf_eos_scf = Dict(dict=wf_eos_scf)

wf_eos = Dict(dict=wf_eos)

calc_eos = {'comp': {'kmax': 3.8,
                     },
            'kpt': {'div1': 4,
                    'div2': 4,
                    'div3': 4
                    }
            }

calc_eos = Dict(dict=calc_eos)

options_eos = {'resources': {'num_machines': 1, 'num_mpiprocs_per_machine': 4, 'num_cores_per_mpiproc': 6},
               'queue_name': 'devel',
               'custom_scheduler_commands': '',
               'max_wallclock_seconds':  1*60*60}

options_eos = Dict(dict=options_eos)

wf_relax = {'film_distance_relaxation': False,
            'force_criterion': 0.049,
            'relax_iter': 5
            }

wf_relax_scf = {'fleur_runmax': 5,
                'use_relax_xml': True,
                'itmax_per_run': 50,
                'alpha_mix': 0.015,
                'relax_iter': 25,
                'force_converged': 0.001,
                'force_dict': {'qfix': 2,
                               'forcealpha': 0.75,
                               'forcemix': 'straight'},
                'inpxml_changes': []
                }

wf_relax = Dict(dict=wf_relax)
wf_relax_scf = Dict(dict=wf_relax_scf)

calc_relax = {'comp': {'kmax': 4.0,
                       },
              'kpt': {'div1': 24,
                      'div2': 20,
                      'div3': 1
                      },
              'atom': {'element': 'Pt',
                       'rmt': 2.2,
                       'lmax': 10,
                       'lnonsph': 6,
                       'econfig': '[Kr] 5s2 4d10 4f14 5p6| 5d9 6s1',
                       },
              'atom2': {'element': 'Fe',
                        'rmt': 2.1,
                        'lmax': 10,
                        'lnonsph': 6,
                        'econfig': '[Ne] 3s2 3p6| 3d6 4s2',
                        },
              }

calc_relax = Dict(dict=calc_relax)

options_relax = {'resources': {'num_machines': 1, 'num_mpiprocs_per_machine': 4, 'num_cores_per_mpiproc': 6},
                 'queue_name': 'devel',
                 'custom_scheduler_commands': '',
                 'max_wallclock_seconds':  1*60*60}

inputs = {
    'eos': {
        'scf': {
            'wf_parameters': wf_eos_scf,
            'calc_parameters': calc_eos,
            'options': options_eos,
            'inpgen': inpgen_code,
            'fleur': fleur_code
        },
        'wf_parameters': wf_eos
    },
    'relax': {
        'scf': {
            'wf_parameters': wf_relax_scf,
            'calc_parameters': calc_relax,
            'options': options_relax,
            'inpgen': inpgen_code,
            'fleur': fleur_code
        },
        'wf_parameters': wf_relax,
    },
    'wf_parameters': wf_para
}

res = submit(FleurCreateMagneticWorkChain, **inputs)



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
In here we put all things util (methods, code snipets) that are often useful, but not yet in AiiDA
itself.
So far it contains:

export_extras
import_extras
delete_nodes (FIXME)
delete_trash (FIXME)
create_group

"""
# TODO import, export of descriptions, and labels...?
import json

from six.moves import input as input_six

from aiida.orm import load_node
from aiida.orm.querybuilder import QueryBuilder
from aiida.orm import Group, Node
from aiida.common.exceptions import NotExistent


def export_extras(nodes, filename='node_extras.txt'):
    """
    Writes uuids and extras of given nodes to a json-file.
    This is useful for import/export because currently extras are lost.
    Therefore this can be used to save and restore the extras via
    :func:`~aiida_fleur.tools.common_aiida.import_extras`.

    :param: nodes: list of AiiDA nodes, pks, or uuids
    :param: filename, string where to store the file and its name

    example use:
    .. code-block:: python

        node_list = [120,121,123,46]
        export_extras(node_list)

    """

    outdict = {}
    for node in nodes:
        if not isinstance(node, Node):  # pk or uuid
            node = load_node(node)

        uuid = node.uuid
        extras_dict = node.extras
        outdict[uuid] = extras_dict

    with open(filename, 'w') as file:
        json.dump(outdict, file, sort_keys=True, indent=4, separators=(',', ': '))


def import_extras(filename):
    """
    Reads in node uuids and extras from a file (most probably generated by
    :func:`~aiida_fleur.tools.common_aiida.export_extras`) and applies them to nodes in the DB.

    This is useful for import/export because currently extras are lost.
    Therefore this can be used to save and restore the extras on the nodes.

    :param: filename, string what file to read from (has to be json format)

    example use:
    import_extras('node_extras.txt')
    """

    all_extras = {}
    with open(filename) as file1:
        try:
            all_extras = json.load(file1)
        except json.JSONDecodeError:
            print('The file has to be loadable by json. i.e json format (which it is not).')

    for uuid, extras in all_extras.items():

        try:
            node = load_node(uuid)
        except NotExistent:
            # Does not exists
            print(f'node with uuid {uuid} does not exist in DB')
            node = None
            continue
        if isinstance(node, Node):
            node.set_extra_many(extras)
        else:
            print('node is not instance of an AiiDA node')
        #print(extras)


'''
# uncommented this, since it is for sure out of date and dangerous to use
# for current version see aiida-core docs
def delete_nodes(pks_to_delete):
    """
    Delete a set of nodes. (From AiiDA cockbook)
    Note: TODO this has to be improved for workchain removal. (checkpoints and co)
    Also you will be backchecked.

    BE VERY CAREFUL!

    .. note::

        The script will also delete
        all children calculations generated from the specified nodes.

    :params pks_to_delete: a list of the PKs of the nodes to delete
    """
    # TODO: CHECK IF THIS IS UP TO DATE!
    from django.db import transaction
    from django.db.models import Q
    from aiida.backends.djsite.db import models

    # Delete also all children of the given calculations
    # Here I get a set of all pks to actually delete, including
    # all children nodes.
    all_pks_to_delete = set(pks_to_delete)
    for pk in pks_to_delete:
        all_pks_to_delete.update(
            models.DbNode.objects.filter(input_links__in=pks_to_delete).values_list('pk', flat=True))

    print(('I am going to delete {} nodes, including ALL THE CHILDREN'
           'of the nodes you specified. Do you want to continue? [y/N]'
           ''.format(len(all_pks_to_delete))))
    answer = input_six()

    if answer.strip().lower() == 'y':
        # Recover the list of folders to delete before actually deleting
        # the nodes.  I will delete the folders only later, so that if
        # there is a problem during the deletion of the nodes in
        # the DB, I don't delete the folders

        # There seem to be no folders in AiiDA 1.0
        #folders = [load_node(pk).folder for pk in all_pks_to_delete]

        with transaction.atomic():
            # Delete all links pointing to or from a given node
            models.DbLink.objects.filter(Q(input__in=all_pks_to_delete) | Q(output__in=all_pks_to_delete)).delete()
            # now delete nodes
            models.DbNode.objects.filter(pk__in=all_pks_to_delete).delete()

        # If we are here, we managed to delete the entries from the DB.
        # I can now delete the folders

        # There seem to be no folders in AiiDA 1.0
        # for f in folders:
        #     f.erase()


def delete_trash():
    """
    This method deletes all AiiDA nodes in the DB, which have a extra trash=True
    And all their children. Could be advanced to a garbage collector.

    Be careful to use it.
    """

    #query db for marked trash
    q = QueryBuilder()
    nodes_to_delete_pks = []

    q.append(Node, filters={'extras.trash': {'==': True}})

    res = q.all()
    for node in res:
        nodes_to_delete_pks.append(node[0].pk)
        print(('pk {}, extras {}'.format(node[0].pk, node[0].extras)))

    #Delete the trash nodes

    print(('deleting nodes {}'.format(nodes_to_delete_pks)))
    delete_nodes(nodes_to_delete_pks)

    return
'''


def create_group(name, nodes, description=None, add_if_exist=False):
    """
    Creates a group for a given node list.

    !!! Now aiida-core has these functionality, use it from there instead!!!
    So far this is only an AiiDA verdi command.

    :params name: string name for the group
    :params nodes: list of AiiDA nodes, pks, or uuids
    :params description: optional string that will be stored as description for the group

    :returns: the group, AiiDa group

    Usage example:

    .. code-block:: python

        group_name = 'delta_structures_gustav'
        nodes_to_group_pks =[2142, 2084]
        create_group(group_name, nodes_to_group_pks,
                     description='delta structures added by hand. from Gustavs inpgen files')

    """
    #from aiida.common import NotExistent

    group, created = Group.objects.get_or_create(label=name)
    if created:
        print(f'Group created with PK={group.pk} and name {group.label}')
    else:
        print(f'Group with name {group.label} and pk {group.pk} already exists.')

        if add_if_exist:
            print(f'Adding nodes to the existing group {group.label}')
        else:
            print(f'Nodes were not added to the existing group {group.label}')
            return

    nodes2 = []
    for node in nodes:
        if not isinstance(node, Node):
            try:
                node = load_node(node)
            except NotExistent:
                print(f'Skipping {node}, it does not exist in the DB')
                continue
        nodes2.append(node)

    group.add_nodes(nodes2)
    print(f'added nodes: {[x.pk for x in nodes2]} to group {group.label} {group.pk}')

    if description:
        group.description = description

    return group


def get_nodes_from_group(group, return_format='uuid'):
    """
    Returns a list of pk or uuid of a nodes in a given group. Since 1.1.0, this function does
    !!! Now aiida-core has these functionality, use it from there instead!!!

    not load a group using the label or any other identification. Use Group.objects.get(filter=ID) to
    pre-load this, available filters are: id, uuid, label, type_string, time, description, user_id.
    """

    if return_format == 'uuid':
        return [x.uuid for x in group.nodes]
    if return_format == 'pk':
        return [x.pk for x in group.nodes]
    else:
        raise ValueError("return_format should be 'uuid' or 'pk'.")



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
This contains code snippets and utility useful for dealing with parameter data nodes
commonly used by the fleur plugin and workflows
"""

import typing as typ
from collections.abc import Mapping


def extract_elementpara(parameter_dict, element):
    """
    :param parameter_dict: python dict, parameter node for inpgen
    :param element: string, i.e 'W'

    :return: python dictionary, parameter node which contains only
                                the atom parameters for the given element
    """
    element_para_dict = {}
    for key, val in parameter_dict.items():
        if 'atom' in key:
            if val.get('element', '') == element:
                element_para_dict[key] = val
        else:
            element_para_dict[key] = val
    return element_para_dict


def dict_merger(dict1, dict2):
    """
    Merge recursively two nested python dictionaries and
    if key is in both digionaries tries to add the entries in both dicts.
    (merges two subdicts, adds lists, strings, floats and numbers together!)

    :param dict1: dict
    :param dict2: dict

    :return dict: Merged dict
    """
    new_dict = dict1.copy()

    if not dict1:
        return dict2
    if not dict2:
        return dict1

    keys1 = list(dict1.keys())
    keys2 = list(dict2.keys())

    # add uncommon
    for key in keys2:
        if key not in keys1:
            new_dict[key] = dict2[key]

    # merge common
    for key, val in dict1.items():
        if isinstance(val, dict):
            new_dict[key] = dict_merger(val, dict2.get(key, {}))
        elif isinstance(val, list):
            new_dict[key] = val + dict2.get(key, [])
        elif isinstance(val, str):
            new_dict[key] = val + dict2.get(key, '')
        elif isinstance(val, int):
            new_dict[key] = val + dict2.get(key, 0)
        elif isinstance(val, float):
            new_dict[key] = val + dict2.get(key, 0.0)
        else:
            print(f"don't know what to do with element : {key}")
    return new_dict


def clean_nones(dict_to_clean):
    """Recursively remove all keys which values are None from a nested dictionary
    return the cleaned dictionary

    :param dict_to_clean: (dict): python dictionary to remove keys with None as value
    :return: dict, cleaned dictionary
    """
    new_dict = {}
    for key, val in dict_to_clean.items():
        if isinstance(val, dict):
            new_val = clean_nones(val)
        else:
            new_val = val
        if new_val is not None:  # currently we keep empty dicts
            new_dict[key] = new_val

    return new_dict


def recursive_merge(left: typ.Dict[str, typ.Any], right: typ.Dict[str, typ.Any]) -> typ.Dict[str, typ.Any]:
    """
    Recursively merge two dictionaries into a single dictionary.

    keys in right override keys in left!


    :param left: first dictionary.
    :param right: second dictionary.
    :return: the recursively merged dictionary.
    """
    for key, value in left.items():
        if key in right:
            if isinstance(value, Mapping) and isinstance(right[key], Mapping):
                right[key] = recursive_merge(value, right[key])

    merged = left.copy()
    merged.update(right)

    return merged


def clear_dict_empty_lists(to_clear_dict):
    """
    Removes entries from a nested dictionary which are empty lists.

    param to_clear_dict dict: python dictionary which should be 'compressed'
    return new_dict dict: compressed python dict version of to_clear_dict

    Hints: recursive
    """
    new_dict = {}
    if not to_clear_dict:
        return new_dict

    if not isinstance(to_clear_dict, dict):
        return to_clear_dict

    for key, value in to_clear_dict.items():
        if value:
            new_value = clear_dict_empty_lists(value)
            if new_value:
                new_dict[key] = new_value
    return new_dict



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
In this module you find the workflow 'FleurEosWorkChain' for the calculation of
of an equation of state
"""
# TODO: print more user info
# allow different inputs, make things optional(don't know yet how)
# half number of iteration if you are close to be converged. (therefore
# one can start with 18 iterations, and if thats not enough run again 9 or something)
import numpy as np

from aiida.orm import load_node
from aiida.orm import Float, StructureData, Dict, List
from aiida.engine import WorkChain, ToContext
from aiida.engine import calcfunction as cf
from aiida.common import AttributeDict

from masci_tools.util.constants import HTR_TO_EV

from aiida_fleur.tools.StructureData_util import rescale, rescale_nowf, is_structure
from aiida_fleur.workflows.scf import FleurScfWorkChain
from aiida_fleur.tools.common_fleur_wf_util import check_eos_energies


class FleurEosWorkChain(WorkChain):
    """
    This workflow calculates the equation of states of a structure.
    Calculates several unit cells with different volumes.
    A Birch_Murnaghan  equation of states fit determines the Bulk modulus and the
    groundstate volume of the cell.

    :params wf_parameters: Dict node, optional 'wf_parameters', protocol specifying parameter dict
    :params structure: StructureData node, 'structure' crystal structure
    :params calc_parameters: Dict node, optional 'calc_parameters' parameters for inpgen
    :params inpgen: Code node,
    :params fleur: Code node,


    :return output_eos_wc_para: Dict node, contains relevant output information.
                                about general succeed, fit results and so on.
    """

    _workflowversion = '0.5.1'

    _default_wf_para = {'points': 9, 'step': 0.005, 'guess': 1.00, 'enforce_same_para': True}
    _default_options = FleurScfWorkChain._default_options

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.expose_inputs(FleurScfWorkChain, namespace='scf', exclude=(
            'structure',
            'remote_data',
            'fleurinp',
        ))
        spec.input('wf_parameters', valid_type=Dict, required=False)
        spec.input('structure', valid_type=StructureData, required=True)

        spec.outline(cls.start, cls.structures, cls.run_first, cls.inspect_first, cls.converge_scf, cls.return_results)

        spec.output('output_eos_wc_para', valid_type=Dict)
        spec.output('output_eos_wc_structure', valid_type=StructureData)

        # exit codes
        spec.exit_code(230, 'ERROR_INVALID_INPUT_PARAM', message='Invalid workchain parameters.')
        spec.exit_code(400,
                       'ERROR_SUB_PROCESS_FAILED',
                       message='At least one of the SCF sub processes did not finish successfully.')

    def start(self):
        """
        check parameters, what condictions? complete?
        check input nodes
        """
        self.report(f'Started eos workflow version {self._workflowversion}')

        self.ctx.last_calc2 = None
        self.ctx.calcs = []
        self.ctx.calcs_future = []
        self.ctx.structures = []
        self.ctx.temp_calc = None
        self.ctx.structures_uuids = []
        self.ctx.scalelist = []
        self.ctx.volume = []
        self.ctx.volume_peratom = {}
        self.ctx.org_volume = -1  # avoid div 0
        self.ctx.labels = []
        self.ctx.successful = True
        self.ctx.info = []
        self.ctx.warnings = []
        self.ctx.errors = []
        # TODO get all successful from convergence, if all True this

        # initialize the dictionary using defaults if no wf paramters are given
        wf_default = self._default_wf_para
        if 'wf_parameters' in self.inputs:
            wf_dict = self.inputs.wf_parameters.get_dict()
        else:
            wf_dict = wf_default

        extra_keys = []
        for key in wf_dict:
            if key not in wf_default:
                extra_keys.append(key)
        if extra_keys:
            error = f'ERROR: input wf_parameters for EOS contains extra keys: {extra_keys}'
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        # extend wf parameters given by user using defaults
        for key, val in wf_default.items():
            wf_dict[key] = wf_dict.get(key, val)
        self.ctx.wf_dict = wf_dict

        self.ctx.points = wf_dict.get('points', 9)
        self.ctx.step = wf_dict.get('step', 0.005)
        self.ctx.guess = wf_dict.get('guess', 1.00)
        self.ctx.enforce_para = wf_dict.get('enforce_same_para', True)

    def structures(self):
        """
        Creates structure data nodes with different Volume (lattice constants)
        """
        points = self.ctx.points
        step = self.ctx.step
        guess = self.ctx.guess
        startscale = guess - (points - 1) / 2 * step

        for point in range(points):
            self.ctx.scalelist.append(startscale + point * step)

        self.report(f'scaling factors which will be calculated:{self.ctx.scalelist}')
        self.ctx.org_volume = self.inputs.structure.get_cell_volume()

        struc_dict = eos_structures(self.inputs.structure, List(list=self.ctx.scalelist))
        # since cf this has to be a dict, we sort to assure ordering of scale
        self.ctx.structures = [struc_dict[key] for key in sorted(struc_dict)]

    def run_first(self):
        """
        Launch the first fleur SCF workchain
        """
        calcs = {}

        i = 0
        struc = self.ctx.structures[i]
        inputs = self.get_inputs_scf_first()
        inputs.structure = struc
        natoms = len(struc.sites)
        label = f'scale_{self.ctx.scalelist[i]}'.replace('.', '_')
        label_c = '|eos| fleur_scf_wc'
        description = f'|FleurEosWorkChain|fleur_scf_wc|{label}, {i}'

        self.ctx.volume.append(struc.get_cell_volume())
        self.ctx.volume_peratom[label] = struc.get_cell_volume() / natoms
        self.ctx.structures_uuids.append(struc.uuid)

        result = self.submit(FleurScfWorkChain, **inputs)
        self.ctx.labels.append(label)
        calcs[label] = result

        return ToContext(**calcs)

    def inspect_first(self):
        """
        Check if the first calculation failed and
        """
        label = self.ctx.labels[0]
        first_scf = self.ctx[label]
        if not first_scf.is_finished_ok:
            self.report('Initial sub process did not finish successfully; aborting the workchain.')
            # return self.exit_codes.ERROR_SUB_PROCESS_FAILED.format(cls=self.inputs.sub_process_class)  # pylint: disable=no-member
            return self.exit_codes.ERROR_SUB_PROCESS_FAILED

        fleurinp = first_scf.outputs.fleurinp
        self.ctx.first_calc_parameters = fleurinp.get_parameterdata()

    def converge_scf(self):
        """
        Launch fleur_scfs from the generated structures
        """
        calcs = {}

        for i, struc in enumerate(self.ctx.structures[1:]):
            inputs = self.get_inputs_scf()
            inputs.structure = struc
            natoms = len(struc.sites)
            label = f'scale_{self.ctx.scalelist[i + 1]}'.replace('.', '_')
            label_c = '|eos| fleur_scf_wc'
            description = f'|FleurEosWorkChain|fleur_scf_wc|{label}, {i+1}'
            #inputs.label = label_c
            #inputs.description = description

            self.ctx.volume.append(struc.get_cell_volume())
            self.ctx.volume_peratom[label] = struc.get_cell_volume() / natoms
            self.ctx.structures_uuids.append(struc.uuid)

            result = self.submit(FleurScfWorkChain, **inputs)
            self.ctx.labels.append(label)
            calcs[label] = result

        return ToContext(**calcs)

    def get_inputs_scf_first(self):
        """
        get and 'produce' the inputs for a scf-cycle
        """
        input_scf = AttributeDict(self.exposed_inputs(FleurScfWorkChain, namespace='scf'))

        return input_scf

    def get_inputs_scf(self):
        """
        get and 'produce' the inputs for a scf-cycle
        """
        input_scf = AttributeDict(self.exposed_inputs(FleurScfWorkChain, namespace='scf'))

        # ensure that all are run with the same FLAPW parameters
        if ('calc_parameters' not in input_scf) and self.ctx.enforce_para:
            # TODO maybe merge with user given calcparameters...
            input_scf['calc_parameters'] = self.ctx.first_calc_parameters

        return input_scf

    def return_results(self):
        """
        return the results of the calculations  (scf workchains) and do a
        Birch-Murnaghan fit for the equation of states
        """
        distancelist = []
        t_energylist = []
        t_energylist_peratom = []
        vol_peratom_success = []
        outnodedict = {}
        natoms = len(self.inputs.structure.sites)
        e_u = 'eV'
        dis_u = 'me/bohr^3'
        for label in self.ctx.labels:
            calc = self.ctx[label]

            if not calc.is_finished_ok:
                message = f'One SCF workflow was not successful: {label}'
                self.ctx.warnings.append(message)
                self.ctx.successful = False
                continue

            try:
                outputnode_scf = calc.outputs.output_scf_wc_para
            except KeyError:
                message = f'One SCF workflow failed, no scf output node: {label}. I skip this one.'
                self.ctx.errors.append(message)
                self.ctx.successful = False
                continue

            outnodedict[label] = outputnode_scf

            outpara = outputnode_scf.get_dict()

            t_e = outpara.get('total_energy', float('nan'))
            e_u = outpara.get('total_energy_units', 'eV')
            if e_u in ['Htr', 'htr']:
                t_e = t_e * HTR_TO_EV
            dis = outpara.get('distance_charge', float('nan'))
            dis_u = outpara.get('distance_charge_units', 'me/bohr^3')
            t_energylist.append(t_e)
            t_energylist_peratom.append(t_e / natoms)
            vol_peratom_success.append(self.ctx.volume_peratom[label])
            distancelist.append(dis)

        not_ok, an_index = check_eos_energies(t_energylist_peratom)

        if not_ok:
            message = f'Abnormality in Total energy list detected. Check entr(ies) {an_index}.'
            hint = ('Consider refining your basis set.')
            self.ctx.info.append(hint)
            self.ctx.warnings.append(message)

        en_array = np.array(t_energylist_peratom)
        vol_array = np.array(vol_peratom_success)

        write_defaults_fit = False
        # TODO: different fits
        if len(en_array):  # for some reason just en_array does not work
            volume, bulk_modulus, bulk_deriv, residuals = birch_murnaghan_fit(en_array, vol_array)

            # something went wrong with the fit
            for i in volume, bulk_modulus, bulk_deriv, residuals:
                if issubclass(type(i), np.complex):
                    write_defaults_fit = True

            if all(i is not None for i in (volume, bulk_modulus, bulk_deriv, residuals)):
                # cast float, because np datatypes are sometimes not serialable
                volume, bulk_modulus = float(volume), float(bulk_modulus)
                bulk_deriv, residuals = float(bulk_deriv), residuals.tolist()

                volumes = self.ctx.volume
                gs_scale = volume * natoms / self.ctx.org_volume
                bulk_modulus = bulk_modulus * 160.217733  # *echarge*1.0e21,#GPa
                if (volume * natoms < volumes[0]) or (volume * natoms > volumes[-1]):
                    warn = ('Groundstate volume was not in the scaling range.')
                    hint = f'Consider rerunning around point {gs_scale}'
                    self.ctx.info.append(hint)
                    self.ctx.warnings.append(warn)
                    # TODO maybe make it a feature to rerun with centered around the gs.
        else:
            write_defaults_fit = True

        if write_defaults_fit:
            volumes = None
            gs_scale = None
            residuals = None
            volume = 0
            bulk_modulus = None
            bulk_deriv = None

        out = {
            'workflow_name': self.__class__.__name__,
            'workflow_version': self._workflowversion,
            'scaling': self.ctx.scalelist,
            'scaling_gs': gs_scale,
            'initial_structure': self.inputs.structure.uuid,
            'volume_gs': volume * natoms,
            'volumes': volumes,
            'volume_units': 'A^3',
            'natoms': natoms,
            'total_energy': t_energylist,
            'total_energy_units': e_u,
            'structures': self.ctx.structures_uuids,
            'calculations': [],  # self.ctx.calcs1,
            'scf_wfs': [],  # self.converge_scf_uuids,
            'distance_charge': distancelist,
            'distance_charge_units': dis_u,
            'nsteps': self.ctx.points,
            'guess': self.ctx.guess,
            'stepsize': self.ctx.step,
            # 'fitresults' : [a, latticeconstant, c],
            # 'fit' : fit_new,
            'residuals': residuals,
            'bulk_deriv': bulk_deriv,
            'bulk_modulus': bulk_modulus,
            'bulk_modulus_units': 'GPa',
            'info': self.ctx.info,
            'warnings': self.ctx.warnings,
            'errors': self.ctx.errors
        }

        if self.ctx.successful:
            self.report('Done, Equation of states calculation complete')
        else:
            self.report('Done, but something went wrong.... Probably some individual calculation failed or'
                        ' a scf-cycle did not reach the desired distance.')

        outnode = Dict(dict=out)
        outnodedict['results_node'] = outnode

        # create links between all these nodes...
        outputnode_dict = create_eos_result_node(**outnodedict)
        outputnode = outputnode_dict.get('output_eos_wc_para')
        outputnode.label = 'output_eos_wc_para'
        outputnode.description = ('Contains equation of states results and information of an FleurEosWorkChain run.')

        returndict = {}
        returndict['output_eos_wc_para'] = outputnode

        outputstructure = outputnode_dict.get('gs_structure', None)
        if outputstructure:
            outputstructure.label = 'output_eos_wc_structure'
            outputstructure.description = ('Structure with the scaling/volume of the lowest total '
                                           'energy extracted from FleurEosWorkChain')

            returndict['output_eos_wc_structure'] = outputstructure

        # create link to workchain node
        for link_name, node in returndict.items():
            self.out(link_name, node)

    def control_end_wc(self, errormsg):
        """
        Controlled way to shutdown the workchain. It will initialize the output nodes
        The shutdown of the workchain will has to be done afterwards
        """
        self.ctx.successful = False
        self.report(errormsg)
        self.ctx.errors.append(errormsg)
        self.return_results()


@cf
def create_eos_result_node(**kwargs):
    """
    This is a pseudo cf, to create the right graph structure of AiiDA.
    This calcfunction will create the output nodes in the database.
    It also connects the output_nodes to all nodes the information comes from.
    This includes the output_parameter node for the eos, connections to run scfs,
    and returning of the gs_structure (best scale)
    So far it is just parsed in as kwargs argument, because we are to lazy
    to put most of the code overworked from return_results in here.
    """
    outdict = {}
    outpara = kwargs.get('results_node', {})
    outdict['output_eos_wc_para'] = outpara.clone()
    # copy, because we rather produce the same node twice
    # then have a circle in the database for now...
    outputdict = outpara.get_dict()
    structure = load_node(outputdict.get('initial_structure'))
    gs_scaling = outputdict.get('scaling_gs', 0)
    if gs_scaling:
        gs_structure = rescale_nowf(structure, Float(gs_scaling))
        outdict['gs_structure'] = gs_structure

    return outdict


@cf
def eos_structures(inp_structure, scalelist):
    """
    Calcfunction, which creates many rescaled StructureData nodes out of a given crystal structure.
    Keeps the provenance in the database

    :param StructureData, a StructureData node
    :param scalelist, AiiDA List, list of floats, scaling factors for the cell

    :returns: dict of New StructureData nodes with rescalled structure, which are linked to input
              Structure
    """
    # we do this in one calcfunction now to store less nodes in the DB
    re_strucs = eos_structures_nocf(inp_structure, scalelist)

    # in AiiDA link labels are always strings, because of namespaces '.' are not allowed.
    # replace '.' by underscore to store floats in link label
    res_new = {}
    for key, struc in re_strucs.items():
        # label already set by rescale_nowf
        struc.description = str(key)
        link_name = f'scale_{key}'.replace('.', '_')
        res_new[link_name] = struc

    return res_new


def eos_structures_nocf(inp_structure, scalelist):
    """
    Creates many rescalled StructureData nodes out of a crystal structure.
    Does NOT keep the provenance in the database.

    :param StructureData, a StructureData node (pk, sor uuid)
    :param scalelist, list of floats, scaling factors for the cell

    :returns: dict of New StructureData nodes with rescalled structure, key=scale
    """
    structure = is_structure(inp_structure)
    if not structure:
        # TODO: log something (test if it gets here at all)
        return None
    re_structures = {}

    for scale in scalelist:
        structure_rescaled = rescale_nowf(structure, scale)  # this is not a cf
        re_structures[scale] = structure_rescaled

    return re_structures


# pylint: disable=invalid-name


def birch_murnaghan_fit(energies, volumes):
    """
    least squares fit of a Birch-Murnaghan equation of state curve. From delta project
    containing in its columns the volumes in A^3/atom and energies in eV/atom
    # The following code is based on the source code of eos.py from the Atomic
    # Simulation Environment (ASE) <https://wiki.fysik.dtu.dk/ase/>.
    :params energies: list (numpy arrays!) of total energies eV/atom
    :params volumes: list (numpy arrays!) of volumes in A^3/atom

    #volume, bulk_modulus, bulk_deriv, residuals = Birch_Murnaghan_fit(data)
    """
    fitdata = np.polyfit(volumes[:]**(-2. / 3.), energies[:], 3, full=True)
    ssr = fitdata[1]
    sst = np.sum((energies[:] - np.average(energies[:]))**2.)
    #print(ssr, sst, energies)
    if sst == 0:
        residuals0 = -1
    else:
        residuals0 = ssr / sst
    deriv0 = np.poly1d(fitdata[0])
    deriv1 = np.polyder(deriv0, 1)
    deriv2 = np.polyder(deriv1, 1)
    deriv3 = np.polyder(deriv2, 1)

    volume0 = 0
    x = 0
    for x in np.roots(deriv1):
        if x > 0 and deriv2(x) > 0:
            volume0 = x**(-3. / 2.)
            break

    if volume0 == 0:
        print('Error: No minimum could be found')
        return None, None, None, None  #exit()

    derivV2 = 4. / 9. * x**5. * deriv2(x)
    derivV3 = (-20. / 9. * x**(13. / 2.) * deriv2(x) - 8. / 27. * x**(15. / 2.) * deriv3(x))
    bulk_modulus0 = derivV2 / x**(3. / 2.)
    bulk_deriv0 = -1 - x**(-3. / 2.) * derivV3 / derivV2

    return volume0, bulk_modulus0, bulk_deriv0, residuals0


def birch_murnaghan(volumes, volume0, bulk_modulus0, bulk_deriv0):
    """
    This evaluates the Birch Murnaghan equation of states
    """
    PV = []
    EV = []
    v0 = volume0
    bm = bulk_modulus0
    dbm = bulk_deriv0

    for vol in volumes:
        pv_val = 3 * bm / 2. * ((v0 / vol)**(7 / 3.) - (v0 / vol)**(5 / 3.)) * \
            (1 + 3 / 4. * (dbm - 4) * ((v0 / vol)**(2 / 3.) - 1))
        PV.append(pv_val)
        ev_val = 9 * bm * v0 / 16. * ((dbm * (v0 / vol)**(2 / 3.) - 1)**(3) * ((v0 / vol)**(2 / 3.) - 1)**2 *
                                      (6 - 4 * (v0 / vol)**(2 / 3.)))
        EV.append(ev_val)
    return EV, PV



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
'''
AiiDA-FLEUR workflows
'''



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
In here we put all things (methods) that are common to workflows AND DO NOT
depend on AiiDA classes, therefore can be used without loading the dbenv.
Util that does depend on AiiDA classes should go somewhere else.
"""

from math import gcd  # pylint: disable=no-name-in-module

from sympy import Symbol


def convert_formula_to_formula_unit(formula):
    """
    Converts a formula to the smallest chemical formula unit
    'Be4W2' -> 'Be2W'
    """

    # get formula dict
    # find greatest common divider of values
    # form formula unit string
    element_count_dict = get_natoms_element(formula)
    nelements = list(element_count_dict.values())
    g = int(nelements[0])
    for a2 in nelements:
        a2 = int(a2)
        g = gcd(g, a2)

    formula_unit_string = ''
    for key, val in element_count_dict.items():
        new_val = int(val / g)
        if new_val == 1:
            new_val = ''
        formula_unit_string = formula_unit_string + f'{key}{new_val}'

    return formula_unit_string


def get_natoms_element(formula):
    """
    Converts 'Be24W2' to {'Be': 24, 'W' : 2}, also BeW to {'Be' : 1, 'W' : 1}
    """

    import re
    elem_count_dict = {}
    elements = re.findall('[A-Z][^A-Z]*', formula)
    #re.split('(\D+)', formula)

    for i, elm in enumerate(elements):
        elem_count = re.findall(r'\D+|\d+\.\d+|\d+', elm)
        #print(elem_count)
        if len(elem_count) == 1:
            elem_count_dict[elem_count[0]] = 1
        else:
            elem_count_dict[elem_count[0]] = float(elem_count[1])

    return elem_count_dict


# test
#get_natoms_element('BeW')
#get_natoms_element('Be2W')


def convert_frac_formula(formula, max_digits=3):
    """
    Converts a formula with fractions to a formula with integer factors only

    Be0.5W0.5 -> BeW

    :param formula: str, crystal formula i.e. Be2W, Be0.2W0.7
    :param max_digits: int default=3, number of digits after which fractions will be cut off
    :returns: string
    """
    form_dict = get_natoms_element(formula)
    formula_int = ''
    for key, val in form_dict.items():
        formula_int = formula_int + key + str(int(val * 10**max_digits))

    return convert_formula_to_formula_unit(formula_int)


#test
#convert_frac_formula('Be0.3W0.7') -> Be3W7
#convert_frac_formula('Be0.5W0.5') -> BeW
#convert_frac_formula('Be3W7') -> Be3W7


def ucell_to_atompr(ratio, formulas, element, error_ratio=None):
    """
    Converts unit cell ratios into atom ratios.

    len(ratio) == len(formulas) (== len(error_ratio))
    ucell_to_atompr([10, 1, 7], ['Be12Ti', 'Be17Ti2', 'Be2'], element='Be', [0.1, 0.1, 0.1])
    """
    import numpy as np

    if error_ratio is None:
        error_ratio = []

    atompro = []
    atompro_err = []
    if len(ratio) != len(formulas):
        return

    n_atoms_formula = []
    for formula in formulas:
        res = get_natoms_element(formula)
        n_atoms_formula.append(res.get(element, 0))

    atompro = np.array(ratio) * np.array(n_atoms_formula)
    total = sum(atompro)
    atompro = atompro / total

    if len(error_ratio):
        atompro_err_t = np.array(error_ratio) * np.array(n_atoms_formula)
        e_sum = np.sqrt(sum(atompro_err_t**2))
        atompro_err = 1 / total * (np.sqrt(atompro_err_t**2 + (atompro * e_sum)**2))

    return atompro, atompro_err


def calc_stoi(unitcellratios, formulas, error_ratio=None):
    """
    Calculate the Stoichiometry with errors from a given unit cell ratio, formulas.

    Example:
    calc_stoi([10, 1, 7], ['Be12Ti', 'Be17Ti2', 'Be2'], [0.1, 0.01, 0.1])
    ({'Be': 12.583333333333334, 'Ti': 1.0}, {'Be': 0.12621369924887876, 'Ti': 0.0012256517540566825})
    calc_stoi([10, 1, 7], ['Be12Ti', 'Be17Ti2', 'Be2'])
    ({'Be': 12.583333333333334, 'Ti': 1.0}, {})
    """
    import numpy as np

    if error_ratio is None:
        error_ratio = []

    stoi = {}
    if len(unitcellratios) != len(formulas):
        return

    errors_stoi = {}
    for i, formula in enumerate(formulas):
        res = get_natoms_element(formula)
        for element, val in res.items():
            stoi_elm = stoi.get(element, 0)
            stoi[element] = stoi_elm + val * unitcellratios[i]
            if len(error_ratio):
                errors = errors_stoi.get(element, 0)
                errors_stoi[element] = errors + val * val * error_ratio[i] * error_ratio[i]

    # make smallest number always one.
    vals = list(stoi.values())
    minv = min(vals)
    keymin = list(stoi.keys())[vals.index(minv)]
    norm_stoi = {}
    for key, val in stoi.items():
        norm_stoi[key] = stoi[key] / minv
        if len(error_ratio):
            errors_stoi[key] = 1 / stoi[keymin] * np.sqrt(errors_stoi[key]**2 +
                                                          (stoi[key] / stoi[keymin] * errors_stoi[keymin])**2)
    return norm_stoi, errors_stoi


def get_atomprocent(formula):
    """
    This converts a formula to a dictionary with element : atomprocent
    example converts 'Be24W2' to {'Be': 24/26, 'W' : 2/26}, also BeW to {'Be' : 0.5, 'W' : 0.5}
    :params: formula: string
    :returns: a dict, element : atomprocent

    # Todo alternative with structuredata
    """
    form_dict_new = {}
    form_dict = get_natoms_element(formula)
    ntotal = sum(form_dict.values())
    for key, val in form_dict.items():
        val_new = float(val) / ntotal
        form_dict_new[key] = val_new
    return form_dict_new


# test
'''
def get_weight_procent(formula):
    """
    This converts a formula to a dictionary with element : weightprocent
    example converts 'Be24W2' to {'Be': , 'W' : }, also BeW to {'Be' : , 'W' : }
    :params: formula: string
    :returns: a dict, element : weightprocent

    # Todo alternative with structuredata
    """

    pass
'''

#def norm_total_energy_peratom(totalenergy, formula)
#def norm_total_energy_perunitcell(totalenergy, )


def determine_formation_energy(struc_te_dict, ref_struc_te_dict):
    """
    This method determines the formation energy.
    E_form =  E(A_xB_y) - x*E(A) - y*E(B)

    :params struc_te_dict: python dictionary in the form of {'formula' : total_energy} for the compound(s)
    :params ref_struc_te_dict: python dictionary in the form of {'formula' : total_energy per atom, or per unit cell} for the elements
                               (if the formula of the elements contains a number the total energy is divided by that number)
    :returns: list of floats, dict {formula : eform, ..} units energy/per atom, energies have some unit as energies given
    """
    #eform_list = []
    eform_dict = {}
    #ref_el = ref_struc_te_dict.keys()
    ref_struc_te_dict_norm = ref_struc_te_dict  #{}
    # assume reference to be normalized

    # normalize reference
    #for key, val in ref_struc_te_dict.iteritems():
    ##    elem_n = get_natoms_element(key)
    #    ref_struc_te_dict_norm[elem_n.keys()[0]] = val / elem_n.values()[0]
    ref_el_norm = list(ref_struc_te_dict_norm.keys())

    for formula, tE in struc_te_dict.items():
        elements_count = get_natoms_element(formula)
        ntotal = float(sum(elements_count.values()))
        print(ntotal)
        eform = tE  #abs(tE)
        for elem, count in elements_count.items():
            if elem in ref_el_norm:
                eform = eform - count * ref_struc_te_dict_norm.get(elem)  #abs(ref_struc_te_dict.get(elem))
            else:
                print('Reference energy missing for element {}. '
                      'You need to provide reference energies for all elements in you compound.'
                      ''.format(elem))
        eform_dict[formula] = eform / ntotal
        #eform_list.append(eform/ntotal)
    return list(eform_dict.values()), eform_dict


# test
#determine_formation_energy({'BeW' : 2, 'Be2W' : 2.5}, {'Be' : 1, 'W' : 1})


def determine_convex_hull(formation_en_grid):
    """
    Wraps the pyhull packge implementing the qhull algo for our purposes.
    For now only for 2D phase diagrams
    Adds the points [1.0, 0.0] and [0.0, 1.0], because in material science these
    are always there.

    :params: formation_en_grid: list of points in phase space [[x, formation_energy]]
    :returns: a hul datatype
    """
    import numpy as np
    #from scipy.spatial import ConvexHull # Buggy in python 3... some ugly segfault
    from pyhull.convex_hull import ConvexHull

    # TODO multi d
    # check if endpoints are in
    if [1.0, 0.0] not in formation_en_grid:
        formation_en_grid.append([1.0, 0.0])
    if [0.0, 0.0] not in formation_en_grid:
        formation_en_grid.append([0.0, 0.0])

    points = np.array(formation_en_grid)
    hull = ConvexHull(points)

    return hull


def inpgen_dict_set_mesh(inpgendict, mesh):
    """
    params: python dict, used for inpgen parameterdata node
    params: mesh either as returned by kpointsdata or tuple of 3 integers

    returns: python dict, used for inpgen parameterdata node
    """
    if len(mesh) == 2:
        kmesh = mesh[0]
    elif len(mesh) == 3:
        kmesh = mesh
    kpt_dict = inpgendict.get('kpt', {})
    kpt_dict['div1'] = kmesh[0]
    kpt_dict['div2'] = kmesh[1]
    kpt_dict['div3'] = kmesh[2]

    inpgendict_new = inpgendict
    inpgendict_new['kpt'] = kpt_dict

    return inpgendict_new


# test
#inpgen_dict_set_mesh(Be_para.get_dict(), mesh)


def powerset(L):
    """
    Constructs the power set, 'potenz Menge' of a given list.

    return list: of all possible subsets
    """
    import itertools
    pset = []
    for n in range(len(L) + 1):
        for sset in itertools.combinations(L, n):
            pset.append(sset)
    return pset


#a = powerset([1, 2, 3, 4])
#a = powerset(['Be', 'Be2W', 'Be12W', 'Be22W', 'W'])
#print(a)
#print(len(a))


def determine_reactions(formula, available_data):
    """
    Determines and balances theoretical possible reaction.
    Stoichiometry 'Be12W', [Be12W, Be2W, Be, W, Be22W] -> [[Be22W+Be2W], [Be12W], [Be12+W],...]

    :params formula: string, given educts (left side of equation)
    :params available_data: list of strings of compounds (products),
                            from which all possibilities will be constructed
    """

    # 1. for each compound try to balance equation
    # 2. for each compound with any other compound in list, try to balance equation
    # 3. for each compound with each two other compounds ... till other compounds
    reactions = []
    constructed_products = []
    pset_available_data = powerset(available_data)  # if len available_data to large cut off?
    for i, dataset in enumerate(pset_available_data):
        productstring = ''
        if len(dataset) < 1:
            continue
        for entry in dataset:
            productstring = productstring + f'{entry}+'

        productstring = productstring[:-1]
        constructed_products.append(productstring)
        pos_reaction = f'{formula}->{productstring}'
        bal_reaction = balance_equation(pos_reaction, allow_negativ=False, allow_zero=False, eval_linear=True)
        # We do not allow zero coefficients of products, because the resulting equation should already be in our list.
        if bal_reaction:
            # TODO post process (i.e are remove 0 compounds)
            reactions.append(bal_reaction)
        else:
            continue
    return reactions


# test
# reac = determine_reactions('Be12W', ['Be12W', 'Be2W', 'Be', 'W', 'Be22W'])
#print(reac ['1*Be12W->1*Be12W', '1*Be12W->1*Be2W+10*Be', '2*Be12W->1*Be2W+1*Be22W',
#             '1*Be12W->12*Be+1*W', '11*Be12W->5*W+6*Be22W'])
#reac = determine_reactions('Be12Ti', ['Be12Ti', 'Be17Ti2', 'BeTi', 'Ti', 'Be', 'Be2Ti', 'Be8Ti4'])
#print(reac ['1*Be12Ti->1*Be12Ti', '2*Be12Ti->1*Be17Ti2+7*Be', '1*Be12Ti->1*BeTi+11*Be',
#             '1*Be12Ti->1*Ti+12*Be', '1*Be12Ti->10*Be+1*Be2Ti', '4*Be12Ti->40*Be+1*Be8Ti4'])


def convert_eq_to_dict(equationstring):
    """
    Converts an equation string to a dictionary
    convert_eq_to_dict('1*Be12Ti->10*Be+1*Be2Ti+5*Be') ->
    {'products': {'Be': 15, 'Be2Ti': 1}, 'educts': {'Be12Ti': 1}}
    """
    eq_dict = {'products': {}, 'educts': {}}
    product_dict = {}
    educt_dict = {}

    eq_split = equationstring.split('->')
    products = eq_split[1].split('+')
    educts = eq_split[0].split('+')

    for product in products:
        p_list = product.split('*')
        product_dict[p_list[-1]] = int(p_list[0]) + product_dict.get(p_list[-1], 0)
    for educt in educts:
        e_list = educt.split('*')
        educt_dict[e_list[-1]] = int(e_list[0]) + educt_dict.get(e_list[-1], 0)

    eq_dict['products'] = product_dict
    eq_dict['educts'] = educt_dict
    return eq_dict


# test convert_eq_to_dict('1*Be12Ti->10*Be+1*Be2Ti+5*Be')
# {'products': {'Be': 15, 'Be2Ti': 1}, 'educts': {'Be12Ti': 1}}


def get_enhalpy_of_equation(reaction, formenergydict):
    """
    calculate the enthalpy per atom of a given reaction from the given data.

    param reaction: string
    param fromenergydict: dictionary that contains the {compound: formationenergy per atom}

    # TODO check if physics is right
    """
    reac_dict = convert_eq_to_dict(reaction)
    educt_energy = 0
    product_energy = 0

    for compound, factor in reac_dict.get('educts', {}).items():
        compound_e = 0
        try:
            compound_e = formenergydict.get(compound, 0)
        except KeyError:
            print(f'Formation energy of compound {compound} not given in {formenergydict}.I abort...')
            compound_e = 0
            # can be that educt side is not a real 'compound' but just a stoichiometry
            # so we give it 0
            #return None
        educt_energy = educt_energy + factor * compound_e

    for compound, factor in reac_dict.get('products', {}).items():
        try:
            compound_e = formenergydict.get(compound)
        except KeyError:
            print(f'Formation energy of compound {compound} not given in {formenergydict}.I abort...')
            compound_e = 0
            return None
        product_energy = product_energy + factor * compound_e

    return educt_energy - product_energy


def balance_equation(equation_string, allow_negativ=False, allow_zero=False, eval_linear=True):
    """
    Method that balances a chemical equation.

    param equation_string: string (with '->')
    param allow_negativ: bool, default False, allows for negative coefficients for the products.

    return string: balanced equation

    balance_equation("C7H16+O2 -> CO2+H2O"))
    balance_equation("Be12W->Be22W+Be12W")
    balance_equation("Be12W->Be12W")

    have to be intergers everywhere in the equation, factors and formulas

    1*C7H16+11*O2 ->7* CO2+8*H2O
    None
    1*Be12W->1*Be12W
    #TODO The solver better then what we need. Currently if system is over
    #"Be12W->Be2W+W+Be" solves to {a: 24, b: -d/2 + 144, c: d/2 - 120}-> FAIL-> None
    # The code fails in the later stage, but this solution should maybe be used.

    code adapted from stack exchange (the messy part):
    https://codegolf.stackexchange.com/questions/8728/balance-chemical-equations
    """
    import sys
    import re
    from sympy.solvers import solve
    from sympy.core.numbers import Rational, Integer
    from collections import defaultdict
    letters = 'abcdefghijklmnopqrstuvwxyz'
    Ls = list(letters)
    eq = equation_string
    Ss, Os, Es, a, i = defaultdict(list), Ls[:], [], 1, 1
    for p in eq.split('->'):
        for k in p.split('+'):
            c = [Ls.pop(0), 1]
            for e, m in re.findall('([A-Z][a-z]?)([0-9]*)', k):
                m = 1 if m == '' else int(m)
                a *= m
                d = [c[0], c[1] * m * i]
                Ss[e][:0], Es[:0] = [d], [[e, d]]
        i = -1
    Ys = {}  # dict((s, eval('Symbol("' + s + '")')) for s in Os if s not in Ls)
    for s in Os:
        if s not in Ls:
            Ys[s] = Symbol(str(s))
    #a = '+'.join('%d*%s' % (c[1], c[0]) for c in Ss[s])
    #print(a)
    Qs = [
        eval(  # pylint: disable=eval-used
            '+'.join('%d*%s' % (c[1], c[0]) for c in Ss[s]), {}, Ys) for s in Ss
    ] + [Ys['a'] - a]
    # FIXME
    #Qs = []
    #for s in Ss:
    #   res1 = ''
    #   for c in s:
    #       prod = '%d*%s' % (c[1], c[0])
    #       res1 = res1 + '+' + prod
    #   Qs.append(res1)
    #Qs + [Ys['a'] - a]

    k = solve(Qs, *Ys)
    if k:  # if a solution is found multiply by gcd
        # TODO? check if solution has linear dependence: and evaluate
        #for c in k.values():
        #    for char in list(letters):
        #        if char in str(c):
        #             pass
        N = []  #[k[Ys[s]]for s in sorted(Ys)]
        rescale_N = False
        denom_list = []
        for s in sorted(Ys):
            n = k[Ys[s]]
            # idea: check if char in n, then linear depended, then
            try:  # since solver gives also a linear depended solution if correct, but code fails then
                if n < 0 and not allow_negativ:  # We allow for 0 but might be an other case to think about
                    return None
            except TypeError:
                return None  # TODO Maybe other return value... maybe list of some values for
                # linear dependencies, d,e,....? also choose them that the value is positive...?
            if n == 0 and not allow_zero:
                return None
            N.append(n)

            # Rationals are a problem in gcd, so we have to get rid of them
            if isinstance(n, Rational):
                rescale_N = True
                denom_list.append(n.as_numer_denom()[1])
        if rescale_N:
            multiplier = 1
            for denom in denom_list:
                multiplier = multiplier * int(denom)
            N = [int(n * multiplier) for n in N]
        g = N[0]
        for a1, a2 in zip(N[0::2], N[0::2]):
            g = gcd(g, a2)
        N = [int(i / g) for i in N]
        pM = lambda c: str(c) + '*'  # if c!=1 else ''
        res = '->'.join('+'.join(pM(N.pop(0)) + str(t) for t in p.split('+')) for p in eq.split('->'))
        return res
    else:
        return None


def check_eos_energies(energylist):
    """
    Checks if there is an abnormality in the total energies from the Equation of states.
    i.e. if one point has a larger energy then its two neighbors

    :param energylist: list of floats
    :returns nnormalies: integer
    """
    # TODO: possible improvements look at all differences, and introduce a threshold
    # Also look at several points...
    abnormalityindexlist = []
    abnormality = False
    i = 0
    for x, y, z in zip(energylist, energylist[1:], energylist[2:]):
        i = i + 1
        if x < y > z:
            abnormality = True
            abnormalityindexlist.append(i)
            print((x, y, z))
            print('annormly detected')

    return abnormality, abnormalityindexlist



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
    In this module you find the workflow 'FleurRelaxWorkChain' for geometry optimization.
"""
import copy
import numpy as np

from aiida.engine import WorkChain, ToContext, while_, if_
from aiida.engine import calcfunction as cf
from aiida.orm import load_node
from aiida.orm import StructureData, Dict
from aiida.common import AttributeDict
from aiida.common.exceptions import NotExistent

from aiida_fleur.workflows.scf import FleurScfWorkChain
from aiida_fleur.workflows.base_fleur import FleurBaseWorkChain
from aiida_fleur.data.fleurinp import FleurinpData
from aiida_fleur.common.constants import BOHR_A
from aiida_fleur.tools.StructureData_util import break_symmetry_wf


class FleurRelaxWorkChain(WorkChain):
    """
    This workflow performs structure optimization.
    """

    _workflowversion = '0.4.0'

    _default_wf_para = {
        'relax_iter': 5,  # Stop if not converged after so many relaxation steps
        'film_distance_relaxation': False,  # Do not relax the z coordinates
        'force_criterion': 0.001,  # Converge the force until lower this value in atomic units
        'run_final_scf': False,  # Run a final scf on the final relaxed structure
        'break_symmetry': False,  # Break the symmetry for the relaxation each atom own type
        'change_mixing_criterion': 0.025,  # After the force is smaller switch mixing scheme
        'atoms_off': [],  # Species to be switched off, '49999' is reserved
        'relaxation_type': 'atoms'  # others include None and maybe in the future volume
        # None would run an scf only
    }

    _default_options = FleurScfWorkChain._default_options

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.expose_inputs(FleurScfWorkChain, namespace='scf')
        spec.expose_inputs(FleurScfWorkChain,
                           namespace='final_scf',
                           exclude=('structure', 'fleur', 'fleurinp', 'remote_data'),
                           namespace_options={
                               'required': False,
                           })
        spec.input('wf_parameters', valid_type=Dict, required=False)

        spec.outline(
            cls.start,
            if_(cls.should_relax)(cls.converge_scf, cls.check_failure, while_(cls.condition)(
                cls.generate_new_fleurinp,
                cls.converge_scf,
                cls.check_failure,
            )),
            cls.get_results_relax,
            if_(cls.should_run_final_scf)(cls.run_final_scf, cls.get_results_final_scf),
            cls.return_results,
        )

        spec.output('output_relax_wc_para', valid_type=Dict)
        spec.output('optimized_structure', valid_type=StructureData)
        spec.expose_outputs(FleurScfWorkChain, namespace='last_scf')

        # exit codes
        spec.exit_code(230, 'ERROR_INVALID_INPUT_PARAM', message='Invalid workchain parameters.')
        spec.exit_code(231, 'ERROR_INPGEN_MISSING', message='If you want to run a final scf inpgen has to be there.')
        spec.exit_code(350, 'ERROR_DID_NOT_RELAX', message='Optimization cycle did not lead to convergence of forces.')
        spec.exit_code(351, 'ERROR_SCF_FAILED', message='SCF Workchains failed for some reason.')
        spec.exit_code(352, 'ERROR_NO_RELAX_OUTPUT', message='Found no relaxed structure info in the output of SCF')
        spec.exit_code(353, 'ERROR_NO_SCF_OUTPUT', message='Found no SCF output')
        spec.exit_code(354, 'ERROR_SWITCH_BFGS', message='Force is small, switch to BFGS')
        spec.exit_code(311,
                       'ERROR_VACUUM_SPILL_RELAX',
                       message='FLEUR calculation failed because an atom spilled to the'
                       'vacuum during relaxation')
        spec.exit_code(313, 'ERROR_MT_RADII_RELAX', message='Overlapping MT-spheres during relaxation.')

    def start(self):
        """
        Retrieve and initialize paramters of the WorkChain, validate inputs
        """
        self.report(f'INFO: Started structure relaxation workflow version {self._workflowversion}\n')

        self.ctx.info = []  # Collects Hints
        self.ctx.warnings = []  # Collects Warnings
        self.ctx.errors = []  # Collects Errors

        # Pre-initialization of some variables
        self.ctx.loop_count = 0  # Counts relax restarts
        self.ctx.forces = []  # Collects forces
        self.ctx.final_cell = None  # The relaxed Bravais matrix
        self.ctx.final_atom_positions = None  # Relaxed atom positions
        self.ctx.pbc = None  # Boundary conditions
        self.ctx.reached_relax = False  # Bool if is relaxed
        self.ctx.switch_bfgs = False  # Bool if BFGS should be switched on
        self.ctx.scf_res = None  # Last scf results
        self.ctx.final_structure = None  # The optimized structure
        self.ctx.total_magnetic_moment = None

        # initialize the dictionary using defaults if no wf paramters are given
        wf_default = copy.deepcopy(self._default_wf_para)
        if 'wf_parameters' in self.inputs:
            wf_dict = self.inputs.wf_parameters.get_dict()
        else:
            wf_dict = wf_default

        extra_keys = []
        for key in wf_dict.keys():
            if key not in wf_default.keys():
                extra_keys.append(key)
        if extra_keys:
            error = f'ERROR: input wf_parameters for Relax contains extra keys: {extra_keys}'
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        # extend wf parameters given by user using defaults
        for key, val in wf_default.items():
            wf_dict[key] = wf_dict.get(key, val)
        self.ctx.wf_dict = wf_dict

        if '49999' in wf_dict['atoms_off']:
            error = '"49999" label for atoms_off is reserved for internal use'
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        # Check if final scf can be run
        run_final = wf_dict.get('run_final_scf', False)
        if run_final:
            # We need inpgen to be there
            input_scf = AttributeDict(self.exposed_inputs(FleurScfWorkChain, namespace='scf'))

            # policy, reuse as much as possible from scf namespace
            input_final_scf = input_scf
            if 'remote_data' in input_final_scf:
                del input_final_scf.remote_data
            if 'structure' in input_final_scf:
                del input_final_scf.structure
            if 'fleurinp' in input_final_scf:
                del input_final_scf.fleurinp
            if 'wf_parameters' in input_final_scf:
                del input_final_scf.wf_parameters

            if 'final_scf' in self.inputs:
                # Will defaults of namespace override other given options?
                input_final_scf_given = AttributeDict(self.exposed_inputs(FleurScfWorkChain, namespace='final_scf'))
                for key, val in input_final_scf_given.items():
                    input_final_scf[key] = val

            self.ctx.input_final_scf = input_final_scf
            if 'inpgen' not in input_scf and 'inpgen' not in input_final_scf:
                self.report('Error: Wrong input: inpgen missing for final scf.')
                return self.exit_codes.ERROR_INPGEN_MISSING

        # initialize contents to avoid access failures
        self.ctx.total_energy_last = None  #total_energy
        self.ctx.total_energy_units = None  #total_energy_units
        self.ctx.final_cell = None
        self.ctx.final_atom_positions = None  #atom_positions
        self.ctx.atomtype_info = None

    def should_relax(self):
        """
        Should we run a relaxation or only a final scf
        This allows to call the workchain to run an scf only and makes
        logic of other higher workflows a lot easier
        """
        relaxtype = self.ctx.wf_dict.get('relaxation_type', 'atoms')
        if relaxtype is None:
            self.ctx.reached_relax = True
            return False
        else:
            return True

    def converge_scf(self):
        """
        Submits :class:`aiida_fleur.workflows.scf.FleurScfWorkChain`.
        """
        inputs = {}
        if self.ctx.loop_count:
            inputs = self.get_inputs_scf()
        else:
            inputs = self.get_inputs_first_scf()
        res = self.submit(FleurScfWorkChain, **inputs)
        return ToContext(scf_res=res)

    def get_inputs_first_scf(self):
        """
        Initialize inputs for the first iteration.
        """
        input_scf = AttributeDict(self.exposed_inputs(FleurScfWorkChain, namespace='scf'))
        input_scf.metadata.label = 'SCF_forces'
        input_scf.metadata.description = 'The SCF workchain converging forces, part of the Relax'

        if self.ctx.wf_dict['break_symmetry']:
            calc_para = None
            if 'calc_parameters' in input_scf:
                calc_para = input_scf.calc_parameters
            # currently we always break the full symmetry
            break_dict = Dict(dict={'atoms': ['all']})  # for provenance
            broken_sys = break_symmetry_wf(input_scf.structure, wf_para=break_dict, parameterdata=calc_para)
            input_scf.structure = broken_sys['new_structure']
            input_scf.calc_parameters = broken_sys['new_parameters']

        if 'wf_parameters' not in input_scf:
            scf_wf_dict = {}
        else:
            scf_wf_dict = input_scf.wf_parameters.get_dict()

        if 'inpxml_changes' not in scf_wf_dict:
            scf_wf_dict['inpxml_changes'] = []

        scf_wf_dict['mode'] = 'force'

        if self.ctx.wf_dict['film_distance_relaxation']:
            scf_wf_dict['inpxml_changes'].append(('set_atomgroup', {
                'attributedict': {
                    'force': {
                        'relaxXYZ': 'FFT'
                    }
                },
                'species': 'all'
            }))

        for specie_off in self.ctx.wf_dict['atoms_off']:
            scf_wf_dict['inpxml_changes'].append(('set_atomgroup_label', {
                'attributedict': {
                    'force': {
                        'relaxXYZ': 'FFF'
                    }
                },
                'atom_label': specie_off
            }))

        scf_wf_dict['inpxml_changes'].append(('set_atomgroup_label', {
            'attributedict': {
                'force': {
                    'relaxXYZ': 'FFF'
                }
            },
            'atom_label': '49999'
        }))

        input_scf.wf_parameters = Dict(dict=scf_wf_dict)

        return input_scf

    def get_inputs_scf(self):
        """
        Initializes inputs for further iterations.
        """
        input_scf = AttributeDict(self.exposed_inputs(FleurScfWorkChain, namespace='scf'))
        if 'structure' in input_scf:
            del input_scf.structure
            del input_scf.inpgen
            del input_scf.calc_parameters

        if 'wf_parameters' not in input_scf:
            scf_wf_dict = {}
        else:
            scf_wf_dict = input_scf.wf_parameters.get_dict()
            if 'inpxml_changes' in scf_wf_dict:
                old_changes = scf_wf_dict['inpxml_changes']
                new_changes = []
                for change in old_changes:
                    if 'shift_value' not in change[0]:
                        new_changes.append(change)
                scf_wf_dict['inpxml_changes'] = new_changes

        scf_wf_dict['mode'] = 'force'
        input_scf.wf_parameters = Dict(dict=scf_wf_dict)

        scf_wc = self.ctx.scf_res
        input_scf.remote_data = scf_wc.outputs.last_calc.remote_folder
        if self.ctx.new_fleurinp:
            input_scf.fleurinp = self.ctx.new_fleurinp

        return input_scf

    def check_failure(self):
        """
        Throws an exit code if scf failed
        """
        try:
            scf_wc = self.ctx.scf_res
        except AttributeError:
            message = 'ERROR: Something went wrong I do not have new atom positions calculation'
            self.control_end_wc(message)
            return self.exit_codes.ERROR_NO_SCF_OUTPUT

        if not scf_wc.is_finished_ok:
            exit_statuses = FleurScfWorkChain.get_exit_statuses(['ERROR_FLEUR_CALCULATION_FAILED'])
            if scf_wc.exit_status == exit_statuses[0]:
                fleur_calc = scf_wc.outputs.last_calc.remote_folder.creator
                if fleur_calc.exit_status == FleurBaseWorkChain.get_exit_statuses(['ERROR_VACUUM_SPILL_RELAX'])[0]:
                    self.control_end_wc('ERROR: Failed due to atom and vacuum overlap')
                    return self.exit_codes.ERROR_VACUUM_SPILL_RELAX
                elif fleur_calc.exit_status == FleurBaseWorkChain.get_exit_statuses(['ERROR_MT_RADII_RELAX'])[0]:
                    self.control_end_wc('ERROR: Failed due to MT overlap')
                    return self.exit_codes.ERROR_MT_RADII_RELAX
            return self.exit_codes.ERROR_SCF_FAILED

    def condition(self):
        """
        Checks if relaxation criteria is achieved.

        :return: True if structure is optimized and False otherwise
        """
        scf_wc = self.ctx.scf_res

        try:
            relax_data = scf_wc.outputs.last_calc.relax_parameters
        except (NotExistent, AttributeError):
            # TODO: throw exit code
            # message = 'ERROR: Did not manage to read the largest force'
            # self.control_end_wc(message)
            # return self.exit_codes.ERROR_RELAX_FAILED
            return False
        else:
            forces_data = relax_data.get_dict()['posforces'][-1]
            all_forces = []
            for force in forces_data:
                all_forces.extend(force[-3:])
            all_forces = [abs(x) for x in all_forces]
            self.ctx.forces.append(max(all_forces))

        largest_now = self.ctx.forces[-1]

        if largest_now < self.ctx.wf_dict['force_criterion']:
            self.report(f'INFO: Structure is converged to the largest force {self.ctx.forces[-1]}')
            self.ctx.reached_relax = True
            return False
        elif largest_now < self.ctx.wf_dict['change_mixing_criterion'] and self.inputs.scf.wf_parameters['force_dict'][
                'forcemix'] == 'straight':
            self.report(f'INFO: Seems it is safe to switch to BFGS. Current largest force: {self.ctx.forces[-1]}')
            self.ctx.switch_bfgs = True
            return False

        self.ctx.loop_count = self.ctx.loop_count + 1
        if self.ctx.loop_count == self.ctx.wf_dict['relax_iter']:
            self.report('INFO: Reached optimization iteration number {}. Largest force is {}, '
                        'force criterion is {}'.format(self.ctx.loop_count + 1, largest_now,
                                                       self.ctx.wf_dict['force_criterion']))
            return False

        self.report('INFO: submit optimization iteration number {}. Largest force is {}, '
                    'force criterion is {}'.format(self.ctx.loop_count + 1, largest_now,
                                                   self.ctx.wf_dict['force_criterion']))

        return True

    def generate_new_fleurinp(self):
        """
        This function fetches relax.xml from the previous iteration and calls
        :meth:`~aiida_fleur.workflows.relax.FleurRelaxWorkChain.analyse_relax()`.
        New FleurinpData is stored in the context.
        """
        # TODO do we loose provenance here, which we like to keep?
        scf_wc = self.ctx.scf_res
        try:
            relax_parsed = scf_wc.outputs.last_calc.relax_parameters
        except NotExistent:
            return self.exit_codes.ERROR_NO_SCF_OUTPUT

        new_fleurinp = self.analyse_relax(relax_parsed)

        self.ctx.new_fleurinp = new_fleurinp

    @staticmethod
    def analyse_relax(relax_dict):
        """
        This function generates a new fleurinp analysing parsed relax.xml from the previous
        calculation.

        **NOT IMPLEMENTED YET**

        :param relax_dict: parsed relax.xml from the previous calculation
        :return new_fleurinp: new FleurinpData object that will be used for next relax iteration
        """
        # TODO: implement this function, now always use relax.xml generated in FLEUR
        should_relax = False
        if should_relax:
            return 1

        return None

    def should_run_final_scf(self):
        """
        Check if a final scf should be run on the optimized structure
        """
        # Since we run the final scf on the relaxed structure
        return all([self.ctx.wf_dict.get('run_final_scf', False), self.ctx.reached_relax])

    def get_inputs_final_scf(self):
        """
        Initializes inputs for final scf on relaxed structure.
        """
        input_scf = AttributeDict(self.exposed_inputs(FleurScfWorkChain, namespace='scf'))
        input_final_scf = self.ctx.input_final_scf

        if 'wf_parameters' not in input_final_scf:
            # use parameters wf para of relax or defaults
            if 'wf_parameters' not in input_scf:
                scf_wf_dict = {}
            else:
                scf_wf_dict = input_scf.wf_parameters.get_dict()
                if 'inpxml_changes' in scf_wf_dict:
                    old_changes = scf_wf_dict['inpxml_changes']
                    new_changes = []
                    for change in old_changes:
                        if 'shift_value' not in change[0]:
                            new_changes.append(change)
                    scf_wf_dict['inpxml_changes'] = new_changes

            scf_wf_dict['mode'] = 'density'
            input_final_scf.wf_parameters = Dict(dict=scf_wf_dict)
        structure = self.ctx.final_structure
        formula = structure.get_formula()
        input_final_scf.structure = structure
        input_final_scf.fleur = input_scf.fleur
        input_final_scf.metadata.label = f'SCF_final_{formula}'
        input_final_scf.metadata.description = ('Final SCF workchain running on optimized structure {}, '
                                                'part of relax workchain'.format(formula))

        return input_final_scf

    def run_final_scf(self):
        """
        Run a final scf for charge convergence on the optimized structure
        """
        self.report('INFO: Running final SCF after relaxation.')
        inputs = {}
        inputs = self.get_inputs_final_scf()
        res = self.submit(FleurScfWorkChain, **inputs)

        return ToContext(scf_final_res=res)

    def get_results_relax(self):
        """
        Generates results of the workchain.
        Creates a new structure data node which is an
        optimized structure.
        """

        if self.ctx.wf_dict.get('relaxation_type', 'atoms') is None:
            input_scf = AttributeDict(self.exposed_inputs(FleurScfWorkChain, namespace='scf'))
            if 'structure' in input_scf:
                structure = input_scf.structure
            elif 'fleurinp' in input_scf:
                structure = input_scf.fleurinp.get_structuredata_ncf()
            else:
                pass
            self.ctx.final_structure = structure
            self.ctx.final_cell = structure.cell
            # The others are already put to None
            return

        try:
            relax_out = self.ctx.scf_res.outputs.last_fleur_calc_output
            retrieved_node = self.ctx.scf_res.outputs.last_calc.retrieved
        except NotExistent:
            return self.exit_codes.ERROR_NO_SCF_OUTPUT

        relax_out = relax_out.get_dict()

        try:
            total_energy = relax_out['energy']
            total_energy_units = relax_out['energy_units']
            atomtype_info = relax_out['relax_atomtype_info']
        except KeyError:
            return self.exit_codes.ERROR_NO_RELAX_OUTPUT

        self.ctx.total_energy_last = total_energy
        self.ctx.total_energy_units = total_energy_units
        self.ctx.atomtype_info = atomtype_info

        fleurinp = FleurinpData(files=['inp.xml', 'relax.xml'], node=retrieved_node)
        structure = fleurinp.get_structuredata_ncf()

        self.ctx.final_structure = structure

    def get_results_final_scf(self):
        """
        Parser some results of final scf
        """

        try:
            scf_out = self.ctx.scf_final_res.outputs.last_fleur_calc_output
        except NotExistent:
            return self.exit_codes.ERROR_NO_SCF_OUTPUT

        scf_out_d = scf_out.get_dict()
        try:
            total_energy = scf_out_d['energy']
            total_energy_units = scf_out_d['energy_units']
        except KeyError:
            self.report('ERROR: Could not parse total energy of final scf run')
            #return self.exit_codes.ERROR_NO_RELAX_OUTPUT

        self.ctx.total_energy_last = total_energy
        self.ctx.total_energy_units = total_energy_units

        if self.ctx.wf_dict.get('relaxation_type', 'atoms') is None:
            # we need this for run through
            self.ctx.scf_res = self.ctx.scf_final_res

        #if jspin ==2
        try:
            total_mag = scf_out_d['total_magnetic_moment_cell']
            self.ctx.total_magnetic_moment = total_mag
        except KeyError:
            self.report('ERROR: Could not parse total magnetic moment cell of final scf run')

    def return_results(self):
        """
        This function stores results of the workchain into the output nodes.
        """
        #TODO maybe we want to have a more detailed array output node with the force and
        # position history of all atoms?
        out = {
            'workflow_name': self.__class__.__name__,
            'workflow_version': self._workflowversion,
            'energy': self.ctx.total_energy_last,
            'energy_units': self.ctx.total_energy_units,
            'info': self.ctx.info,
            'warnings': self.ctx.warnings,
            'errors': self.ctx.errors,
            'force': self.ctx.forces,
            'force_iter_done': self.ctx.loop_count,
            # uuids in the output are bad for caching should be avoided,
            # instead better return the node.
            'last_scf_wc_uuid': self.ctx.scf_res.uuid,
            'total_magnetic_moment_cell': self.ctx.total_magnetic_moment,
            'total_magnetic_moment_cell_units': 'muBohr'
        }
        outnode = Dict(dict=out)

        con_nodes = {}
        try:
            relax_out = self.ctx.scf_res.outputs.last_fleur_calc_output
        except NotExistent:
            relax_out = None
        if relax_out is not None:
            con_nodes['last_fleur_calc_output'] = relax_out

        if all([self.ctx.wf_dict.get('run_final_scf', False), self.ctx.reached_relax]):
            try:
                scf_out = self.ctx.scf_final_res.outputs.last_fleur_calc_output
            except NotExistent:
                scf_out = None
            if relax_out is not None:
                con_nodes['last_scf__output'] = scf_out

        # TODO: for a trajectory output node all corresponding nodes have to go into
        # con_nodes

        if self.ctx.final_structure is not None:
            outdict = create_relax_result_node(output_relax_wc_para=outnode,
                                               optimized_structure=self.ctx.final_structure,
                                               **con_nodes)
        else:
            outdict = create_relax_result_node(output_relax_wc_para=outnode, **con_nodes)

        #Expose the outputs of the last scf calculation
        self.out_many(self.exposed_outputs(self.ctx.scf_res, FleurScfWorkChain, namespace='last_scf'))

        # return output nodes
        for link_name, node in outdict.items():
            self.out(link_name, node)

        if self.ctx.switch_bfgs:
            return self.exit_codes.ERROR_SWITCH_BFGS
        if not self.ctx.reached_relax:
            return self.exit_codes.ERROR_DID_NOT_RELAX

    def control_end_wc(self, errormsg):
        """
        Controlled way to shutdown the workchain. It will initialize the output nodes
        The shutdown of the workchain will has to be done afterwards.
        """
        self.report(errormsg)
        self.ctx.errors.append(errormsg)
        self.return_results()


@cf
def create_relax_result_node(**kwargs):
    """
    This calcfunction assures the right provenance (additional links)
    for ALL result nodes it takes any nodes as input
    and return a special set of nodes.
    All other inputs will be connected in the DB to these ourput nodes
    """
    outdict = {}
    for key, val in kwargs.items():
        if key == 'output_relax_wc_para':  # should always be present
            outnode = val.clone()  # dublicate node instead of circle (keep DAG)
            outnode.label = 'output_relax_wc_para'
            outnode.description = ('Contains results and information of an FleurRelaxWorkChain run.')
            outdict['output_relax_wc_para'] = outnode

        if key == 'optimized_structure':
            structure = val.clone()  # dublicate node instead of circle (keep DAG)
            structure.label = 'optimized_structure'
            structure.description = ('Relaxed structure result of an FleurRelaxWorkChain run.')
            outdict['optimized_structure'] = structure

    return outdict



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
This module contains the FleurBaseWorkChain.
FleurBaseWorkChain is a workchain that wraps the submission of
the FLEUR calculation. Inheritance from the BaseRestartWorkChain
allows to add scenarios to restart a calculation in an
automatic way if an expected failure occurred.
"""
from aiida import orm
from aiida.common import AttributeDict
from aiida.engine import while_
from aiida.engine.processes.workchains import BaseRestartWorkChain
from aiida.engine.processes.workchains.utils import process_handler, ProcessHandlerReport

from aiida_fleur.tools.common_fleur_wf import optimize_calc_options
from aiida_fleur.calculation.fleur import FleurCalculation


class FleurBaseWorkChain(BaseRestartWorkChain):
    """Workchain to run a FLEUR calculation with automated error handling and restarts"""
    _workflowversion = '0.2.0'
    _process_class = FleurCalculation

    @classmethod
    def define(cls, spec):
        super().define(spec)

        spec.expose_inputs(FleurCalculation, exclude=('metadata.options',))
        spec.input('options', valid_type=orm.Dict, help='Optional parameters to set up computational details.')
        spec.input('description', valid_type=str, required=False, non_db=True, help='Calculation description.')
        spec.input('label', valid_type=str, required=False, non_db=True, help='Calculation label.')
        spec.input(
            'add_comp_para',
            valid_type=orm.Dict,
            default=lambda: orm.Dict(dict={
                'only_even_MPI': False,
                'max_queue_nodes': 20,
                'max_queue_wallclock_sec': 86400
            }),
            help='Gives additional control over computational parameters'
            'only_even_MPI: set to true if you want to suppress odd number of MPI processes in parallelisation.'
            'This might speedup a calculation for machines having even number of sockets per node.'
            'max_queue_nodes: maximal number of nodes allowed on the remote machine. Used only to automatically solve some FLEUR failures.'
            'max_queue_wallclock_sec: maximal wallclock time allowed on the remote machine. Used only to automatically solve some FLEUR failures.'
        )

        spec.outline(
            cls.setup,
            cls.validate_inputs,
            while_(cls.should_run_process)(
                cls.run_process,
                cls.inspect_process,
            ),
            cls.results,
        )

        spec.expose_outputs(FleurCalculation)
        spec.output('final_calc_uuid', valid_type=orm.Str, required=False)

        spec.exit_code(311,
                       'ERROR_VACUUM_SPILL_RELAX',
                       message='FLEUR calculation failed because an atom spilled to the'
                       'vacuum during relaxation')
        spec.exit_code(313, 'ERROR_MT_RADII_RELAX', message='Overlapping MT-spheres during relaxation.')
        spec.exit_code(388, 'ERROR_TIME_LIMIT_NO_SOLUTION', message='Computational resources are not optimal.')
        spec.exit_code(389, 'ERROR_MEMORY_ISSUE_NO_SOLUTION', message='Computational resources are not optimal.')
        spec.exit_code(390, 'ERROR_NOT_OPTIMAL_RESOURCES', message='Computational resources are not optimal.')
        spec.exit_code(399,
                       'ERROR_SOMETHING_WENT_WRONG',
                       message='FleurCalculation failed and FleurBaseWorkChain has no strategy '
                       'to resolve this')

    def validate_inputs(self):
        """
        Validate inputs that might depend on each other and cannot be validated by the spec.
        Also define dictionary `inputs` in the context, that will contain the inputs for the
        calculation that will be launched in the `run_calculation` step.
        """
        self.ctx.inputs = AttributeDict(self.exposed_inputs(FleurCalculation))

        self.ctx.max_queue_nodes = self.inputs.add_comp_para['max_queue_nodes']
        self.ctx.max_queue_wallclock_sec = self.inputs.add_comp_para['max_queue_wallclock_sec']

        input_options = self.inputs.options.get_dict()
        self.ctx.optimize_resources = input_options.pop('optimize_resources', True)
        self.ctx.inputs.metadata.options = input_options

        if 'description' in self.inputs:
            self.ctx.inputs.metadata.description = self.inputs.description
        else:
            self.ctx.inputs.metadata.description = ''
        if 'label' in self.inputs:
            self.ctx.inputs.metadata.label = self.inputs.label
        else:
            self.ctx.inputs.metadata.label = ''

        if not self.ctx.optimize_resources:
            self.ctx.can_be_optimised = False  # set this for handlers to not change resources
            return

        resources_input = self.ctx.inputs.metadata.options['resources']
        try:
            self.ctx.num_machines = int(resources_input['num_machines'])
            self.ctx.num_mpiprocs_per_machine = int(resources_input['num_mpiprocs_per_machine'])
        except KeyError:
            self.ctx.can_be_optimised = False
            self.report('WARNING: Computation resources were not optimised.')
        else:
            try:
                self.ctx.num_cores_per_mpiproc = int(resources_input['num_cores_per_mpiproc'])
                self.ctx.use_omp = True
                self.ctx.suggest_mpi_omp_ratio = self.ctx.num_mpiprocs_per_machine / self.ctx.num_cores_per_mpiproc
            except KeyError:
                self.ctx.num_cores_per_mpiproc = 1
                self.ctx.use_omp = False
                self.ctx.suggest_mpi_omp_ratio = 1

            try:
                self.check_kpts()
                self.ctx.can_be_optimised = True
            except Warning:
                self.report('ERROR: Not optimal computational resources.')
                return self.exit_codes.ERROR_NOT_OPTIMAL_RESOURCES

    def check_kpts(self):
        """
        This routine checks if the total number of requested cpus
        is a factor of kpts and makes an optimisation.

        If suggested number of num_mpiprocs_per_machine is 60% smaller than
        requested, it throws an exit code and calculation stop withour submission.
        """
        fleurinp = self.ctx.inputs.fleurinpdata
        machines = self.ctx.num_machines
        mpi_proc = self.ctx.num_mpiprocs_per_machine
        omp_per_mpi = self.ctx.num_cores_per_mpiproc
        only_even_MPI = self.inputs.add_comp_para['only_even_MPI']
        try:
            adv_nodes, adv_mpi_tasks, adv_omp_per_mpi, message = optimize_calc_options(machines,
                                                                                       mpi_proc,
                                                                                       omp_per_mpi,
                                                                                       self.ctx.use_omp,
                                                                                       self.ctx.suggest_mpi_omp_ratio,
                                                                                       fleurinp,
                                                                                       only_even_MPI=only_even_MPI)
        except ValueError as exc:
            raise Warning('Not optimal computational resources, load less than 60%') from exc

        self.report(message)

        self.ctx.inputs.metadata.options['resources']['num_machines'] = adv_nodes
        self.ctx.inputs.metadata.options['resources']['num_mpiprocs_per_machine'] = adv_mpi_tasks
        if self.ctx.use_omp:
            self.ctx.inputs.metadata.options['resources']['num_cores_per_mpiproc'] = adv_omp_per_mpi
            if 'environment_variables' in self.ctx.inputs.metadata.options:
                self.ctx.inputs.metadata.options['environment_variables']['OMP_NUM_THREADS'] = str(adv_omp_per_mpi)
            else:
                self.ctx.inputs.metadata.options['environment_variables'] = {}
                self.ctx.inputs.metadata.options['environment_variables']['OMP_NUM_THREADS'] = str(adv_omp_per_mpi)

    @process_handler(priority=1,
                     exit_codes=[
                         FleurCalculation.exit_codes.ERROR_FLEUR_CALC_FAILED,
                         FleurCalculation.exit_codes.ERROR_MT_RADII,
                         FleurCalculation.exit_codes.ERROR_NO_RETRIEVED_FOLDER,
                         FleurCalculation.exit_codes.ERROR_OPENING_OUTPUTS,
                         FleurCalculation.exit_codes.ERROR_NO_OUTXML,
                         FleurCalculation.exit_codes.ERROR_XMLOUT_PARSING_FAILED,
                         FleurCalculation.exit_codes.ERROR_RELAX_PARSING_FAILED,
                         FleurCalculation.exit_codes.ERROR_MISSING_DEPENDENCY,
                     ])
    def _handle_general_error(self, calculation):
        """
        Calculation failed for unknown reason.
        """
        self.ctx.restart_calc = calculation
        self.ctx.is_finished = True
        self.report('Calculation failed for a reason that can not be resolved automatically')
        self.results()
        return ProcessHandlerReport(True, self.exit_codes.ERROR_SOMETHING_WENT_WRONG)

    @process_handler(priority=48, exit_codes=FleurCalculation.exit_codes.ERROR_DROP_CDN)
    def _handle_dirac_equation(self, calculation):
        """
        Sometimes relaxation calculation fails with Diraq problem which is usually caused by
        problems with reusing charge density. In this case we resubmit the calculation, dropping the input cdn.
        """

        # try to drop remote folder and see if it helps
        is_fleurinp_from_relax = False
        if 'fleurinpdata' in self.ctx.inputs:
            if 'relax.xml' in self.ctx.inputs.fleurinpdata.files:
                is_fleurinp_from_relax = True

        if 'parent_folder' in self.ctx.inputs and is_fleurinp_from_relax:
            del self.ctx.inputs.parent_folder
            self.ctx.restart_calc = None
            self.ctx.is_finished = False
            self.report('Calculation seems to fail due to corrupted charge density (can happen'
                        'during relaxation). I drop cdn from previous step')
            return ProcessHandlerReport(True)

        self.ctx.restart_calc = calculation
        self.ctx.is_finished = True
        self.report('Can not drop charge density. If I drop the remote folder, there will be no inp.xml')
        self.results()
        return ProcessHandlerReport(True, self.exit_codes.ERROR_SOMETHING_WENT_WRONG)

    @process_handler(priority=52, exit_codes=FleurCalculation.exit_codes.ERROR_VACUUM_SPILL_RELAX)
    def _handle_vacuum_spill_error(self, calculation):
        """
        Calculation failed for unknown reason.
        """

        self.ctx.restart_calc = calculation
        self.ctx.is_finished = True
        self.report('FLEUR calculation failed because an atom spilled to the vacuum during'
                    'relaxation. Can be fixed via RelaxBaseWorkChain.')
        self.results()
        return ProcessHandlerReport(True, self.exit_codes.ERROR_VACUUM_SPILL_RELAX)

    @process_handler(priority=51, exit_codes=FleurCalculation.exit_codes.ERROR_MT_RADII_RELAX)
    def _handle_mt_relax_error(self, calculation):
        """
        Calculation failed for unknown reason.
        """
        self.ctx.restart_calc = calculation
        self.ctx.is_finished = True
        self.report('FLEUR calculation failed due to MT overlap. Can be fixed via RelaxBaseWorkChain')
        self.results()
        return ProcessHandlerReport(True, self.exit_codes.ERROR_MT_RADII_RELAX)

    @process_handler(priority=50, exit_codes=FleurCalculation.exit_codes.ERROR_NOT_ENOUGH_MEMORY)
    def _handle_not_enough_memory(self, calculation):
        """
        Calculation failed due to lack of memory.
        Probably works for JURECA only, has to be tested for other systems.
        """

        if self.ctx.can_be_optimised:
            self.ctx.restart_calc = None
            self.ctx.is_finished = False
            self.report('Calculation failed due to lack of memory, I resubmit it with twice larger'
                        ' amount of computational nodes and smaller MPI/OMP ratio')

            # increase number of nodes
            propose_nodes = self.ctx.num_machines * 2
            if propose_nodes > self.ctx.max_queue_nodes:
                propose_nodes = self.ctx.max_queue_nodes
            self.ctx.num_machines = propose_nodes

            self.ctx.suggest_mpi_omp_ratio = self.ctx.suggest_mpi_omp_ratio / 2
            self.check_kpts()

            if 'settings' not in self.ctx.inputs:
                self.ctx.inputs.settings = {}
            else:
                self.ctx.inputs.settings = self.ctx.inputs.settings.get_dict()
            self.ctx.inputs.settings.setdefault('remove_from_remotecopy_list', [])
            if 'mixing_history*' not in self.ctx.inputs.settings['remove_from_remotecopy_list']:
                self.ctx.inputs.settings['remove_from_remotecopy_list'].append('mixing_history*')
            return ProcessHandlerReport(True)
        else:
            self.ctx.restart_calc = calculation
            self.ctx.is_finished = True
            self.report('I am not allowed to optimize your settings. Consider providing at least'
                        'num_machines and num_mpiprocs_per_machine')
            self.results()
            return ProcessHandlerReport(True, self.exit_codes.ERROR_MEMORY_ISSUE_NO_SOLUTION)

    @process_handler(priority=47, exit_codes=FleurCalculation.exit_codes.ERROR_TIME_LIMIT)
    def _handle_time_limits(self, calculation):
        """
        If calculation fails due to time limits, we simply resubmit it.
        """
        from aiida.common.exceptions import NotExistent

        # if previous calculation failed for the same reason, do not restart
        try:
            prev_calculation_remote = calculation.get_incoming().get_node_by_label('parent_folder')
            prev_calculation_status = prev_calculation_remote.get_incoming().all()[-1].node.exit_status
            if prev_calculation_status in FleurCalculation.get_exit_statuses(['ERROR_TIME_LIMIT']):
                self.ctx.is_finished = True
                return ProcessHandlerReport(True)
        except NotExistent:
            pass

        self.report('FleurCalculation failed due to time limits, I restart it from where it ended')

        # increase wallclock time
        propose_wallclock = self.ctx.inputs.metadata.options['max_wallclock_seconds'] * 2
        if propose_wallclock > self.ctx.max_queue_wallclock_sec:
            propose_wallclock = self.ctx.max_queue_wallclock_sec
        self.ctx.inputs.metadata.options['max_wallclock_seconds'] = propose_wallclock

        # increase number of nodes
        propose_nodes = self.ctx.num_machines * 2
        if propose_nodes > self.ctx.max_queue_nodes:
            propose_nodes = self.ctx.max_queue_nodes
        self.ctx.num_machines = propose_nodes

        remote = calculation.get_outgoing().get_node_by_label('remote_folder')

        # resubmit providing inp.xml and cdn from the remote folder
        self.ctx.is_finished = False
        check_remote = False

        if 'fleurinpdata' in self.ctx.inputs:
            modes = self.ctx.inputs.fleurinpdata.get_fleur_modes()
            if not (modes['force_theorem'] or modes['dos'] or modes['band']):
                # in modes listed above it makes no sense copying cdn.hdf
                self.ctx.inputs.parent_folder = remote
                del self.ctx.inputs.fleurinpdata
                check_remote = True
        else:
            # it is harder to extract modes in this case - simply try to reuse cdn.hdf and hope it works
            self.ctx.inputs.parent_folder = remote
            check_remote = True

        if check_remote:
            #If no charge density file is available to restart from the calculation will except
            #with a not nice error message. So we try to catch these cases to produce a nice error message
            retrieved_filenames = calculation.get_outgoing().get_node_by_label('retrieved').list_object_names()
            if all(file not in retrieved_filenames for file in (
                    'cdn_last.hdf',
                    'cdn1',
            )):
                self.report(
                    'FleurCalculation failed due to time limits and no charge density file is available. Aborting!')
                return ProcessHandlerReport(True, self.exit_codes.ERROR_TIME_LIMIT_NO_SOLUTION)

        return ProcessHandlerReport(True)
